(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Loop
imports SubtermCriterion
begin

  
lemma loop_imp_not_SN : assumes steps: "(s,C\<langle>s \<cdot> \<sigma>\<rangle>) \<in> (rstep R)^+"
  shows "\<not> SN_elt (rstep R) s"
proof
  let ?R = "op \<rhd> \<union> rstep R"
  assume "SN_elt (rstep R) s"
  from SN_elt_rstep_imp_SN_elt_supt_union_rstep[OF this]
  have SN: "SN_elt ?R s" .
  from sub_closed_union[OF supt_sub_closed sub_closed_rstep]
  have sub: "sub_closed ?R" .
  have steps: "(s, C\<langle>s \<cdot> \<sigma>\<rangle>) \<in> ?R^+" using trancl_mono[OF steps] by auto
  have "(s, s \<cdot> \<sigma>) \<in> ?R^+"
  proof (cases C)
    case Hole
    with steps  show ?thesis by simp
  next
    case (More f bef D aft)
    with nectxt_imp_suptp[of C _ "s \<cdot> \<sigma>"] have "C\<langle>s \<cdot> \<sigma>\<rangle> \<rhd> s \<cdot> \<sigma>" by simp 
    hence "(C\<langle>s \<cdot> \<sigma>\<rangle>, s \<cdot> \<sigma>) \<in> ?R" unfolding supt_def by auto
    with steps show ?thesis by (rule trancl_into_trancl)
  qed
  from stable_loop_imp_not_SN[OF sub this] SN 
  show False ..
qed    


fun
  infinite_dpp :: "('f, 'v) dpp \<Rightarrow> bool"
where
  "infinite_dpp (P, R) = ((\<not> SN (rstep R)) \<or> (\<exists>s t \<sigma>. ichain (P, R) s t \<sigma>))"

fun infinite_dpp_alt :: "('f,'v)dpp \<Rightarrow> bool"
where "infinite_dpp_alt (P,R) = (\<not> SN (rrstep P \<union> rstep R))"


lemma infinite_dpp_alt: "infinite_dpp = infinite_dpp_alt"
proof (intro ext, clarify)
  fix P R
  show "infinite_dpp (P,R) = infinite_dpp_alt (P,R)"
  proof (cases "SN (rstep R)")
    case False
    with SN_subset[of "rrstep P \<union> rstep R" "rstep R"]
    show ?thesis by auto
  next
    case True note snR = this
    let ?R = "rstep R"
    let ?P = "rrstep P"
    let ?RP = "?P \<union> ?R"
    show ?thesis
    proof
      assume "infinite_dpp (P,R)"
      with snR obtain s t \<sigma> where ichain: "ichain (P,R) s t \<sigma>" by auto
      {
        fix i
        let ?i = "Suc i"
        from ichain have st:  "(s ?i, t ?i) \<in> P" by simp
        hence step: "(s ?i \<cdot> \<sigma> ?i, t ?i \<cdot> \<sigma> ?i) \<in> ?RP" unfolding rrstep_def rstep_r_p_s_def by auto
        from ichain have ts: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?R^*" by simp
        with rtrancl_mono[of ?R ?RP] have steps: "(t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?RP^*" by auto
        from step steps have "(t i \<cdot> \<sigma> i, t (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?RP^+" by (simp only: trancl_to_rtrancl, auto)
      }
      hence "\<not> SN (?RP^+)" unfolding SN_defs by force
      with SN_imp_SN_trancl have "\<not> SN ?RP" by auto
      thus "infinite_dpp_alt (P,R)" by simp
    next
      assume "infinite_dpp_alt (P,R)"
      then obtain f where steps: "\<And> i. (f i, f (Suc i)) \<in> ?RP" by (simp, unfold SN_defs, auto)
      show "infinite_dpp (P,R)"
      proof (cases "\<exists> i. \<forall> j \<ge> i. (f j, f (Suc j)) \<in> ?R")
        case True
        then obtain i where rsteps: "\<And> j. j \<ge> i \<Longrightarrow> (f j, f (Suc j)) \<in> ?R" by auto
        obtain g where g: "\<And> j. g j = f (j + i)" by auto
        from rsteps have "\<And> j. (g j, g (Suc j)) \<in> ?R" using g by auto
        hence "\<not> SN ?R" unfolding SN_defs by force
        with snR show ?thesis by simp
      next
        case False
        hence "\<forall> i. \<exists> j \<ge> i. (f j, f (Suc j)) \<notin> ?R" by simp
        with steps have Psteps: "\<forall> i. \<exists> j \<ge> i. (f j, f (Suc j)) \<in> ?P" by blast                
        obtain p where p: "\<And> j. p j = ((f j, f (Suc j)) \<in> ?P)" by auto
        { 
          fix i
          assume "\<not> p i"
          hence "(f i, f (Suc i)) \<notin> ?P" unfolding p by auto
          with steps have "(f i, f (Suc i)) \<in> ?R" by auto
        } note p_false = this
        {
          fix i
          assume "p i"
          hence "(f i, f (Suc i)) \<in> ?P" unfolding p by auto
          hence "\<exists> s t \<sigma> . (s,t) \<in> P \<and> s \<cdot> \<sigma> = f i \<and> t \<cdot> \<sigma> = f (Suc i)"
            unfolding rrstep_def rstep_r_p_s_def by auto
        } note p_true = this
        from Psteps have inf_p: "\<forall> i. \<exists> j \<ge> i. p j" unfolding p by auto
        hence ih: "infinite_hits p" by (unfold_locales, simp)
        obtain p_index where "p_index = infinite_hits.index p" by simp
        with infinite_hits.index_p[OF ih] infinite_hits.index_ordered[OF ih] infinite_hits.index_not_p_between[OF ih] 
        have p_index: "\<And> i. p (p_index i) \<and> p_index i < p_index (Suc i) \<and> (\<forall> j. p_index i < j \<and> j < p_index (Suc i) \<longrightarrow> \<not> p j)" by auto
        let ?prop = "\<lambda> s t \<sigma> i. ((s,t) \<in> P \<and> s \<cdot> \<sigma> = f (p_index i) \<and> t \<cdot> \<sigma> = f (Suc (p_index i)) \<and> (f (Suc (p_index i)), f (p_index (Suc i))) \<in> ?R^*)"
        {
          fix i
          let ?pi = "p_index i"
          let ?psi = "p_index (Suc i)"
          from p_index p_true[of ?pi] obtain s t \<sigma> 
            where st: "(s,t) \<in> P" and s: "s \<cdot> \<sigma> = f ?pi" and t: "t \<cdot> \<sigma> = f (Suc ?pi)" by auto
          from p_index have isi: "?pi < ?psi" by auto
          obtain pi psi where pi: "pi = ?pi" and psi: "psi = ?psi" by auto
          with p_index[of i] p_false have inter: "\<And> j. pi < j \<and> j < psi \<Longrightarrow> (f j, f (Suc j)) \<in> ?R" by auto
          from pi isi psi have pisi: "pi < psi" by simp                      
          {
            fix n
            assume "Suc n \<le> psi - pi"
            hence "(f (Suc pi), f (Suc (n + pi))) \<in> ?R^*"
            proof (induct n, simp)
              case (Suc n)
              hence steps: "(f (Suc pi), f (Suc (n+pi))) \<in> ?R^*" by simp
              have "(f (Suc (n+pi)), f (Suc (Suc n + pi))) \<in> ?R"
                using inter[of "Suc n + pi"] Suc(2) by auto
              with steps show ?case by simp
            qed
          }
          from this[of "psi - pi - 1"] pisi have 
            "(f (Suc pi), f psi) \<in> ?R^*" by simp
          with pi psi have rsteps: "(f (Suc ?pi), f ?psi) \<in> ?R^*" by simp
          from rsteps st s t have "\<exists> s t \<sigma>. ?prop s t \<sigma> i" by auto
        }
        hence "\<forall> i. \<exists> s t \<sigma>. ?prop s t \<sigma> i" by simp
        from choice[OF this] obtain s where "\<forall> i. \<exists> t \<sigma>. ?prop (s i) t \<sigma> i" by blast
        from choice[OF this] obtain t where "\<forall> i. \<exists> \<sigma>. ?prop (s i) (t i) \<sigma> i" by blast
        from choice[OF this] obtain \<sigma> where "\<And> i. ?prop (s i) (t i) (\<sigma> i) i" by blast
        hence "ichain (P,R) s t \<sigma>" by simp
        thus ?thesis by (simp del: ichain.simps, blast)
      qed
    qed
  qed
qed          


lemma loop_imp_infinite : assumes "(s,s \<cdot> \<sigma>) \<in> (rrstep P \<union> rstep R)^+"
  shows "infinite_dpp (P,R)"
proof -
  obtain PR where PR: "PR = rrstep P \<union> rstep R" by auto
  from sub_closed_union[OF sub_closed_rrstep sub_closed_rstep] PR have sub: "sub_closed PR" by simp
  from PR assms have steps: "(s,s \<cdot> \<sigma>) \<in> PR^+" by simp
  from stable_loop_imp_not_SN[OF sub steps] have "\<not> SN PR" unfolding SN_def by auto
  thus ?thesis
    by (simp add: infinite_dpp_alt PR)
qed

end
