(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory ProofChecker
imports SubtermCriterionImpl DepGraph RedPairImpl DpTransImpl NonTermination
  SemanticLabeling RootLabeling StringReversal
begin

text {* 
   A polynomial order is represented by a list of function symbol assignments where 
   every function symbol obtains a pair of a number n and a list of numbers [n1,..,nk].
   The corresponding polynomial interpretation is as follows:
   Pol(f(x1,..,xl)) = n + sum_i ni * xi
        where i ranges from 1 to min(k,l)
*}

types
  ('f, 'l, 'v) trsLL = "(('f, 'l) lab, 'v) rules"
  ('f, 'l, 'v) rseqL = "((('f, 'l) lab, 'v) rule \<times> (('f, 'l) lab, 'v) rseq) list"

datatype ('f, 'l, 'v) "DPProof" =
  PisEmpty |
  RedPairProc "('f,'l)lab redpair" "('f,'l,'v)trsLL"  "('f,'l,'v)DPProof" |
  RedPairUrProc "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v) DPProof" |
  DepGraphProc "(('f,'l,'v)DPProof option \<times> ('f,'l,'v)trsLL) list" |
  MonoRedPairProc "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  MonoRedPairUrProc "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  UnlabProc "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  SemlabProc "('f,'l)sl_inter" "('f,nat,'l)label" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  RootlabProc "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  SubtermCriterionProc "('f, 'l) lab spL" "('f, 'l, 'v) rseqL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof" |
  SizeChangeSubTerm "((('f,'l)lab,'v)rule \<times> ((nat\<times>nat)list \<times> (nat\<times>nat)list))list" |
  SizeChangeRedPair "('f,'l)lab redpair" "('f,'l,'v)trsLL" "((('f,'l)lab,'v)rule \<times> ((nat\<times>nat)list \<times> (nat\<times>nat)list))list" |
  FccProc "('f, 'l)lab" "(('f,'l)lab,'v)ctxt list" "('f,'l,'v)trsLL" "('f,'l,'v)trsLL" "('f,'l,'v)DPProof"

definition
  indent :: "shows \<Rightarrow> shows"
where
  "indent p s \<equiv> flat_map (\<lambda>c. (if c = newline then newline # '' '' else [c])) (p s)"

types
  ('f, 'l, 'v) ruleMap = "(('f, 'l) lab \<Rightarrow> ('f, 'l, 'v) trsLL) option"

(* TODO: move to Monadic_Check *)
(* allows to modify the index during the check *)
definition
  check_allm_gen_index :: "('a \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> ('a \<Rightarrow> nat \<Rightarrow> 'e check) \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> 'e check"
where
  "check_allm_gen_index g f n xs \<equiv> snd (foldl (\<lambda>(i, m) x. (g x i, m \<guillemotright> f x i)) (n, OK) xs)"

lemma foldl_error:
  "snd (foldl (\<lambda>(i, m) x . (g x i, m \<guillemotright> f x i)) (n, error e) xs) = error e"
by (induct xs arbitrary: n) auto

lemma isOK_check_allm_gen_index[simp]:
  assumes "isOK (check_allm_gen_index g f n xs)"
  shows "\<forall>x\<in>set xs. \<exists>i. isOK (f x i)"
using assms proof (induct xs arbitrary: n)
  case Nil thus ?case by simp
next
  case (Cons x xs)
  show ?case
  proof (cases "isOK (f x n)")
    case True
    hence "\<exists>i. isOK (f x i)" by auto
    with True Cons show ?thesis unfolding check_allm_gen_index_def by auto
  next
    case False
    then obtain e where "f x n = error e" by (cases "f x n") auto
    with foldl_error[of g f _ e] and Cons show ?thesis
      unfolding check_allm_gen_index_def by auto
  qed
qed

lemma check_allm_gen_index[fundef_cong]:
  fixes f :: "'a \<Rightarrow> nat \<Rightarrow> 'e check"
  assumes "\<And>x n. x \<in> set xs \<Longrightarrow> g x n = g' x n"
    and "\<And>x n. x \<in> set xs \<Longrightarrow> f x n = f' x n"
  shows "check_allm_gen_index g f n xs = check_allm_gen_index g' f' n xs"
proof -
  {
    fix n m
    from assms
      have "foldl (\<lambda>(i, m) x. (g x i, m \<guillemotright> f x i)) (n, m) xs
        = foldl (\<lambda>(i, m) x. (g' x i, m \<guillemotright> f' x i)) (n, m) xs"
      by (induct xs arbitrary: n m) auto
  }
  thus ?thesis unfolding check_allm_gen_index_def by simp
qed

function
  checkDPProof ::
    "shows \<Rightarrow> ('f, 'l, string) trsLL \<Rightarrow> ('f, 'l, string) trsLL \<Rightarrow> nat
     \<Rightarrow> ('f, 'l, string) ruleMap
     \<Rightarrow> ('f, 'l, string) ruleMap \<Rightarrow> ('f::show, 'l::show, string) DPProof
     \<Rightarrow> shows check" 
where 
  "checkDPProof i P R m rm rrm PisEmpty = debug (i []) ''P is empty''
     (case P of
       Nil \<Rightarrow> return ()
     | Cons _ _ \<Rightarrow> 
       error (i +@+ '': P is not empty in the following DP-problem '' +#+
         shows_nl +@+ shows_dpp P R))" |
  "checkDPProof i P R m rm rrm (RedPairProc redp rP prf) = debug (i []) ''RedPairProc'' (do {
         create_redpair_proc redp P rP R 
           <+? (\<lambda>s. i +@+ '': error when applying the reduction pair processor to switch from''
             +#+ shows_nl +@+ shows_dpp P R +@+ shows_nl +@+
             ''to the DP problem'' +#+ shows_nl +@+ shows_dpp rP R +@+ shows_nl +@+ s);
         checkDPProof (i +@+ shows_string ''.1'') rP R m rm rrm prf
           <+? (\<lambda>s. i +@+ '': error below the reduction pair processor'' +#+ shows_nl
             +@+ indent s)
   })" |
  "checkDPProof i P R m rm rrm (RedPairUrProc redp rP ur prf)  =
     debug (i []) ''RedPairProc (usable rules + af)'' (do {
       create_redpair_ur_af_proc redp P rP R rm ur
         <+? (\<lambda>s. i +@+ '': error when applying the reduction pair '' +#+
           ''processor (with usable rules and af) to switch from '' +#+ shows_nl +@+
           shows_dpp P R +@+ shows_nl +@+ ''to the DP problem '' +#+ shows_nl +@+
           shows_dpp rP R +@+ shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') rP R m rm rrm prf
         <+? (\<lambda>s. i +@+ '': error below the reduction pair processor'' +#+
           shows_nl +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (DepGraphProc edpts) =
     debug (i []) ''DepGraphProc'' (do {
       check_dep_graph_proc P (map (\<lambda>x.
         ((case fst x of Some _ \<Rightarrow> True | None \<Rightarrow> False), snd x)) edpts) rm rrm
         <+? (\<lambda>s. i +@+ '': error while trying to perform SCC-decomposition  on '' +#+
           shows_nl +@+ shows_dpp P R +@+ shows_nl +@+ s);
        check_allm_gen_index
          (\<lambda>rp j. (case fst rp of Some prf \<Rightarrow> Suc j | None \<Rightarrow> j))
          (\<lambda>rp j.
            (case fst rp of
              Some prf \<Rightarrow> checkDPProof (i +@+ ''.'' +#+ shows j) (snd rp) R m rm rrm prf
            | None \<Rightarrow> OK))
          1 edpts
          <+? (\<lambda>s. i +@+ '': error below the dependency graph processor'' +#+ shows_nl
            +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (MonoRedPairProc redp rP rR prf)  =
     debug (i []) ''MonoRedPairProc'' (do {
       create_mono_redpair_proc redp P rP R rR
         <+? (\<lambda>s. i +@+ '': error when applying the monotonic reduction pair processor ''
           +#+ ''to switch from '' +#+ shows_nl +@+ shows_dpp P R +@+ shows_nl +@+
           ''to the DP problem '' +#+ shows_nl +@+ shows_dpp rP rR +@+ shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') rP rR m
         (computeRuleMap rR) (computeRuleMap (reverse_trs rR)) prf
         <+? (\<lambda>s. i +@+ '': error below the monotonic reduction pair processor'' +#+
           shows_nl +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (MonoRedPairUrProc redp rP rR ur prf) =
     debug (i []) ''MonoRedPairProc (usable rules)'' (do {
       create_mono_redpair_ur_proc redp P rP R rm rR ur
         <+? (\<lambda>s. i +@+ '': error when applying the monotonic reduction pair '' +#+
           ''processor (with usable rules) to switch from '' +#+ shows_nl +@+ shows_dpp P R
           +@+ shows_nl +@+ ''to the DP problem'' +#+ shows_nl +@+ shows_dpp rP rR +@+
           shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') rP rR m
         (computeRuleMap rR) (computeRuleMap (reverse_trs rR)) prf
         <+? (\<lambda>s. i +@+ '': error below the monotonic reduction pair '' +#+
           ''(with usable rules) processor'' +#+ shows_nl +@+ indent s)
     })" |
  "checkDPProof i P R (Suc m) rm rrm (UnlabProc ulP ulR prf) =
     debug (i []) ''UnlabProc'' (do {
       check_unlab_proc P R ulP ulR;
       checkDPProof (i +@+ shows_string ''.1'') ulP ulR m
         (computeRuleMap ulR) (computeRuleMap (reverse_trs ulR)) prf
         <+? (\<lambda>s. i +@+ '': error below the unlabeling processor'' +#+ shows_nl +@+ indent s)
     })" |
  "checkDPProof i P R 0 rm rrm (UnlabProc ulP ulR prf) =
     error (''unlabeling not allowed due to label-depth of 0'' +#+ shows_nl)" |
  "checkDPProof i P R m rm rrm (SemlabProc sli L lP lR prf) =
     debug (i []) ''SemlabProc'' (do {
       create_sem_lab_proc sli L P R lP lR
         <+? (\<lambda>s. i +@+ '': error when applying the semantic labeling processor'' +#+
           shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') lP lR (Suc m)
         (computeRuleMap lR) (computeRuleMap (reverse_trs lR)) prf
         <+? (\<lambda>s. i +@+ '': error below the semantic labeling processor'' +#+ shows_nl
           +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (RootlabProc lP lR prf) =
     debug (i []) ''RootlabProc'' (do {
       rl_proc P R lP lR
         <+? (\<lambda>s. i +@+ '': error when applying the root labeling processor'' +#+ shows_nl
           +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') lP lR (Suc m)
         (computeRuleMap lR) (computeRuleMap (reverse_trs lR)) prf
           <+? (\<lambda>s. i +@+ '': error below the root-labeling processor'' +#+ shows_nl
             +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (SubtermCriterionProc sp rseq rP prf) =
     debug (i []) ''SubtermProc'' (do {
       create_subterm_criterion_proc sp rseq P rP R rm
         <+? (\<lambda>s. i +@+ '': error when applying the subterm criterion processor''
           +#+ shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') rP R m rm rrm prf
         <+? (\<lambda>s. i +@+ '': error below the subterm criterion processor'' +#+ shows_nl
           +@+ indent s)
     })" |
  "checkDPProof i P R m rm rrm (SizeChangeSubTerm graphs) =
     debug (i []) ''SizeChangeSubTerm'' (check_sct_subterm_proc P graphs rm rrm)" |
  "checkDPProof i P R m rm rrm (SizeChangeRedPair redp ur graphs) =
     debug (i []) ''SizeChangeRedPair'' (create_sct_ur_af_proc redp P rm rrm ur graphs)" |
  "checkDPProof i P R m rm rrm (FccProc f fcs rP rR prf) =
     debug (i []) ''FccProc'' (do {
       fcc_proc P R f fcs rP rR
         <+? (\<lambda>s. i +@+ '': error when applying the closure under flat contexts'' +#+
           shows_nl +@+ s);
       checkDPProof (i +@+ shows_string ''.1'') rP rR 0
         (computeRuleMap rR) (computeRuleMap (reverse_trs rR)) prf
         <+? (\<lambda>s. i +@+ '': error below flat context closure processor'' +#+ shows_nl
           +@+ indent s)
     })"
by (pat_completeness) auto

termination "checkDPProof"
proof 
  let ?R = "measures [(\<lambda> (i,P,R,m,rm,rrm,prf). size prf)]"
  show "wf ?R" by auto
next
  fix i P R m rm rrm redp rP "prf" 
  show "((i +@+ shows_string ''.1'', rP, R, m, rm, rrm, prf), i, P, R, m, rm, rrm, RedPairProc redp rP prf) \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm redp rP ur "prf"
  show "((i +@+ shows_string ''.1'', rP, R, m, rm, rrm, prf), i, P, R, m, rm, rrm, RedPairUrProc redp rP ur prf) \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i and P R::"('f::show,'l::show,string)trsLL" and m rm rrm and  
    edpts :: "(('f,'l,string)DPProof option \<times> ('f,'l,string)trsLL) list" and n x and "prf"::"('f,'l,string)DPProof"
  assume elem: "x \<in> set edpts" and "fst x = Some prf"
  hence "(option_size size) (fst x) = Suc (size prf)" by auto
  hence "prod_size (option_size size) (list_size
                (prod_size (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0)) 
                (Term.term.term_size (lab_size (\<lambda>x. 0) (\<lambda>x. 0)) (\<lambda>x. 0)))) x > size prf" 
    (is "?m x > size prf")
    by (cases x, auto)   
  with elem have "size prf < ((list_size ?m) edpts)" 
  proof (induct edpts, simp)
    case (Cons y ys)
    thus ?case by (cases "x = y", auto)
  qed
  with elem show "((i +@+ ''.'' +#+ shows n, snd x, R, m, rm, rrm, prf), i, P, R, m, rm, rrm, DepGraphProc edpts) \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm redp rP rR "prf"
  show "((i +@+ shows_string ''.1'', rP, rR, m, computeRuleMap rR,
         computeRuleMap (reverse_trs rR), prf),
        i, P, R, m, rm, rrm, MonoRedPairProc redp rP rR prf)
       \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm redp rP rR ur "prf"
  show "((i +@+ shows_string ''.1'', rP, rR, m, computeRuleMap rR,
         computeRuleMap (reverse_trs rR), prf),
        i, P, R, m, rm, rrm, MonoRedPairUrProc redp rP rR ur prf)
       \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm ulP ulR "prf" 
  show " ((i +@+ shows_string ''.1'', ulP, ulR, m, computeRuleMap ulR,
         computeRuleMap (reverse_trs ulR), prf),
        i, P, R, Suc m, rm, rrm, UnlabProc ulP ulR prf)
       \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm sli L lP lR "prf"
  show "((i +@+ shows_string ''.1'', lP, lR, Suc m, computeRuleMap lR,
         computeRuleMap (reverse_trs lR), prf),
        i, P, R, m, rm, rrm, SemlabProc sli L lP lR prf)
       \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prf). size prf]" by auto
next
  fix i P R m rm rrm lP lR "prf"
  show "((i +@+ shows_string ''.1'',lP,lR,Suc m,computeRuleMap lR,computeRuleMap(reverse_trs lR),prf),
        i,P,R,m,rm,rrm,RootlabProc lP lR prf)
        \<in> measures [\<lambda>(i,P,R,m,rm,rrm,prf). size prf]" by auto
next
  fix i P R m rm rrm sp rseq rP "prf"
  show "((i +@+ shows_string ''.1'',rP,R,m,rm,rrm,prf),i,P,R,m, rm,rrm,
    SubtermCriterionProc sp rseq rP prf)
    \<in> measures [\<lambda>(i,P,R,m,rm,rrm,prf). size prf]" by auto
next
  fix i P R m rm rrm f fcs rP rR prof
  show "((i +@+ shows_string ''.1'', rP, rR, 0, computeRuleMap rR, computeRuleMap (reverse_trs rR), prof),
    i, P, R, m, rm, rrm, FccProc f fcs rP rR prof)
    \<in> measures [\<lambda>(i, P, R, m, rm, rrm, prof). size prof]" by auto
qed

lemma checkDPProofSoundI:
  fixes P::"('f::show,'l::show,string)trsLL"
  shows "isOK(checkDPProof i P R m rm rrm prf) \<and>
    rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R) \<longrightarrow> finite_lab_dpp (set P, set R) m"
proof (induct rule:checkDPProof.induct[where P = "\<lambda> i P R m rm rrm prf. isOK(checkDPProof i P R m rm rrm prf) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R) \<longrightarrow> finite_lab_dpp (set P, set R) m"])
  case (1 i P R m rm rrm)  thus ?case unfolding finite_lab_dpp_def by (cases P, auto)
next
  case (2 i P R m rm rrm redp rP prof) 
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (RedPairProc redp rP prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule apply_ci_onlyDP_lab_proc[where rP = rP],  rule create_redpair_proc) 
      from rm 2 show "finite_lab_dpp (set rP, set R) m" by force
    next
      from rm 2 show "isOK (create_redpair_proc redp P rP R)" by auto
    qed
  }
  thus ?case by blast
next
  case (3 i P R m rm rrm redp rP ur prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (RedPairUrProc redp rP ur prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule apply_ci_onlyDP_lab_proc[where rP = rP], rule create_redpair_ur_af_proc)
      from rm 3 show "isOK (create_redpair_ur_af_proc redp P rP R (computeRuleMap R) ur)" by auto
    next
      from rm 3 show "finite_lab_dpp (set rP, set R) m" by force
    qed
  }
  thus ?case by blast
next
  case (4 i P R m rm rrm edpts) 
  let ?mc = "\<lambda> edpts. (map (\<lambda> x. ((case fst x of Some _ \<Rightarrow> True | None \<Rightarrow> False), snd x)) edpts)"
  let ?edps = "?mc edpts"
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (DepGraphProc edpts)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule dep_graph_lab_proc, auto)
      from 4 rm show "isOK(check_dep_graph_proc P ?edps (computeRuleMap R) (computeRuleMap (reverse_trs R)))" by (auto)
    next
      fix a rp
      assume elem: "(True, rp) \<in> set ?edps"
      from 4 rm have "isOK (check_allm_gen_index
        (\<lambda>rp j. case fst rp of Some prf \<Rightarrow> Suc j | None \<Rightarrow> j)
        (\<lambda>rp j.
          (case fst rp of
            Some prf \<Rightarrow> checkDPProof (i +@+ ''.'' +#+ shows j) (snd rp) R m rm rrm prf
          | None \<Rightarrow> OK))
        1 edpts)" (is "isOK (check_allm_gen_index ?g ?f 1 edpts)") by auto
      hence "\<forall>prfrP \<in> set edpts. \<exists>n. isOK (?f prfrP n)" by (rule isOK_check_allm_gen_index)
      with 4 rm have fin: "\<forall> prf rP. (Some prf, rP) \<in> set edpts \<longrightarrow> finite_lab_dpp (set rP, set R) m" by auto
      from elem obtain sub where "set sub \<subseteq> set edpts \<and> (True, rp) \<in> set (?mc sub)" by blast
      thus "finite_lab_dpp (set rp, set R) m"
      proof (induct sub, simp)
	case (Cons edp subb)
	with fin show ?case 
	  by (cases "edp", cases "fst edp", simp, cases "snd edp = rp", auto)
      qed
    qed
  }
  thus ?case by blast
next
  case (5 i P R m rm rrm redp rP rR prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (MonoRedPairProc redp rP rR prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule apply_ci_lab_proc[where rP = rP], rule create_mono_redpair_proc)
      from rm 5 show "isOK (create_mono_redpair_proc redp P rP R rR)" by auto
    next
      from rm 5 show "finite_lab_dpp (set rP, set rR) m" by force
    qed
  }
  thus ?case by blast
next
  case (6 i P R m rm rrm redp rP rR ur prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (MonoRedPairUrProc redp rP rR ur prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule apply_ci_lab_proc[where rP = rP], rule create_mono_redpair_ur_proc)
      from rm 6 show "isOK (create_mono_redpair_ur_proc redp P rP R (computeRuleMap R) rR ur)" by auto
    next
      from rm 6 show "finite_lab_dpp (set rP, set rR) m" by force
    qed
  }
  thus ?case by blast
next
  case (7 i P R m rm rrm ulP ulR prof)
  {
    assume rm: "isOK(checkDPProof i P R (Suc m) rm rrm (UnlabProc ulP ulR prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) (Suc m)"
    proof (rule apply_sound_lab_proc[where rP = ulP and rR = ulR], rule check_unlab_proc_sound)
      from rm 7 show "finite_lab_dpp (set ulP, set ulR) m" by (force simp: Let_def)
    next
      from rm 7 show "Suc m = Suc m \<and> isOK(check_unlab_proc P R ulP ulR)" by auto
    qed
  }
  thus ?case by blast
next
  case (8 i P R rm rrm ulP ulR prof)
  thus ?case by auto
next
  case (9 i P R m rm rrm sli L lP lR prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (SemlabProc sli L lP lR prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m" 
    proof (rule apply_sound_lab_proc[where rP = lP], rule create_sem_lab_proc_sound)
      from rm 9 show "finite_lab_dpp (set lP, set lR) (Suc m)" by force
    next
      from rm 9 show "Suc m = Suc m \<and> isOK (create_sem_lab_proc sli L P R lP lR)" by auto
    qed
  }
  thus ?case by blast
next
  case (10 i P R m rm rrm lP lR prof)
  {
    assume rm: "isOK(checkDPProof i P R m rm rrm (RootlabProc lP lR prof)) \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap(reverse_trs R)"
    have "finite_lab_dpp(set P,set R) m"
    proof (rule apply_sound_lab_proc[where rP = lP],rule rl_proc_sound)
      from rm 10 show "finite_lab_dpp(set lP,set lR) (Suc m)" by force
    next
      from rm 10 show "Suc m = Suc m \<and> isOK(rl_proc P R lP lR)" by auto
    qed
  }
  thus ?case by blast
next
  case (11 i P R m rm rrm sp rseq rP prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (SubtermCriterionProc sp rseq rP prof))
      \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp(set P,set R) m"
    proof (rule apply_ci_onlyDP_lab_proc[where rP = rP], rule subterm_criterion_proc_ci)
      from rm 11 show "isOK(create_subterm_criterion_proc sp rseq P rP R (computeRuleMap R))" by auto
    next
      from rm 11 show "finite_lab_dpp(set rP,set R) m" by force
    qed
  }
  thus ?case by blast
next 
  case (12 i P R m rm rrm graphs)
  {
    assume rm: "isOK(checkDPProof i P R m rm rrm (SizeChangeSubTerm graphs))
      \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp(set P, set R) m"
    proof (rule apply_ci_onlyDP_lab_proc[OF check_sct_subterm_proc_sound, of "[]"])
      show "finite_lab_dpp (set [], set R) m" unfolding finite_lab_dpp_def by auto
    next
      from rm 12 show "isOK(check_sct_subterm_proc P graphs (computeRuleMap R) (computeRuleMap (reverse_trs R)))" by force
    qed
  }
  thus ?case by blast
next
  case (13 i P R m rm rrm redp ur graphs)
  {
    assume rm: "isOK(checkDPProof i P R m rm rrm (SizeChangeRedPair redp ur graphs))
      \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp(set P, set R) m"
    proof (rule apply_ci_onlyDP_lab_proc[OF create_sct_ur_af_proc, of "[]"])
      show "finite_lab_dpp (set [], set R) m" unfolding finite_lab_dpp_def by auto
    next
      from rm 13 show "isOK(create_sct_ur_af_proc redp P (computeRuleMap R) (computeRuleMap (reverse_trs R)) ur graphs)" by force
    qed
  }
  thus ?case by blast
next
  case (14 i P R m rm rrm f fcs rP rR prof)
  {
    assume rm: "isOK (checkDPProof i P R m rm rrm (FccProc f fcs rP rR prof))
      \<and> rm = computeRuleMap R \<and> rrm = computeRuleMap (reverse_trs R)"
    have "finite_lab_dpp (set P, set R) m"
    proof (rule apply_sound_proc, rule fcc_proc_sound)
      from rm show "isOK (fcc_proc P R f fcs rP rR)" by auto
    next
      from 14 rm show "finite_lab_dpp (set rP, set rR) 0" by force
    qed
  }
  thus ?case by blast
qed

declare checkDPProof.simps[simp del]

lemma checkDPProofSound: "isOK(checkDPProof i P R m (computeRuleMap R) (computeRuleMap (reverse_trs R)) prf) \<Longrightarrow> finite_lab_dpp (set P, set R) m"
  using checkDPProofSoundI by blast

datatype ('f,'l,'v)"TRSProof" =
  DPTrans "(('f,'l)lab shp,'v) rules"  "(('f,'l)lab shp,'l,'v) DPProof" |
  RuleRemoval "('f,'l)lab redpair" "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  Unlab "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  Semlab "('f,'l)sl_inter" "('f,nat,'l)label" "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  Rootlab "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  Fcc "(('f,'l)lab,'v)ctxt list" "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  StringReversal "('f,'l,'v)trsLL" "('f,'l,'v)TRSProof" |
  RisEmpty

fun
  checkTRSProof ::
    "shows \<Rightarrow> ('f::show, 'l::show, string) trsLL \<Rightarrow> nat \<Rightarrow> ('f, 'l, string) TRSProof
     \<Rightarrow> shows check"
where
  "checkTRSProof i r m (DPTrans P prf) = debug (i []) ''DP trans'' ((\<lambda>IR. ((\<lambda>ILR. do {
     check_DP r P
       <+? (\<lambda>s. i +@+ '': error when switching from the TRS'' +#+ shows_nl +@+ shows_trs r
         +@+ shows_nl +@+ ''to the initial DP problem'' +#+ shows_dpp P IR +@+ shows_nl +@+ s);
     checkDPProof (i +@+ shows_string ''.1'') (init_lab_rules P) ILR 0
       (computeRuleMap ILR) (computeRuleMap (reverse_trs ILR)) prf
       <+? (\<lambda>s. i +@+ '': error below switch to dependency pairs'' +#+ shows_nl +@+ indent s)
   })
   (init_lab_rules IR)))
   (ID_list r))" |
  "checkTRSProof i r m (RuleRemoval rp rR prf) = debug (i []) ''rule removal'' (do {
     create_rule_removal_proc rp r rR
       <+? (\<lambda>s. i +@+ '': error in rule removal when switching from the TRS'' +#+ shows_nl
         +@+ shows_trs r +@+ shows_nl +@+ ''to the TRS'' +#+ shows_nl +@+ shows_trs rR
         +@+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') rR m prf
       <+? (\<lambda>s. i +@+ '': error below the rule removal processor'' +#+ shows_nl +@+ indent s)
   })" |
  "checkTRSProof i r (Suc m) (Unlab uR prf) = debug (i []) ''unlabel'' (do {
     unlab_trs_tt r uR
       <+? (\<lambda>s. i +@+ '': error in unlabeling of TRS when switching from the TRS''
         +#+ shows_nl +@+ shows_trs r +@+ shows_nl +@+ ''to the TRS'' +#+ shows_nl +@+
         shows_trs uR +@+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') uR m prf
       <+? (\<lambda>s. i +@+ '': error below the unlabeling'' +#+ shows_nl +@+ indent s)
   })" |
  "checkTRSProof i r 0 (Unlab uR prf) =
     error (''unlabeling not allowed for label depth 0'' +#+ shows_nl)" |
  "checkTRSProof i r m RisEmpty =
     debug (i []) ''R is empty'' (check (r = [])
       (i +@+ '': R is not empty, but it is the TRS'' +#+ shows_trs r +@+ shows_nl))" |
  "checkTRSProof i R m (Semlab sli L lR prf) = debug (i []) ''Semlab'' (do {
     create_sem_lab_tt sli L R lR
       <+? (\<lambda>s. i +@+ '': error when applying semantic labeling'' +#+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') lR (Suc m) prf
       <+? (\<lambda>s. i +@+ '': error below semantic labeling'' +#+ shows_nl +@+ indent s)
   })" |
  "checkTRSProof i R m (Rootlab lR prf) = debug (i []) ''Rootlab'' (do {
     rl_tt R lR
       <+? (\<lambda>s. i +@+ '': error when applying root-labeling'' +#+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') lR (Suc m) prf
       <+? (\<lambda>s. i +@+ '': error below root-labeling'' +#+ shows_nl +@+ indent s)
   })" |
  "checkTRSProof i R m (Fcc fcs rR prf) = debug (i []) ''Fcc'' (do {
     fcc_tt R fcs rR
       <+? (\<lambda>s. i +@+ '': error when applying flat context closure'' +#+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') rR 0 prf
       <+? (\<lambda>s. i +@+ '': error below flat context closure'' +#+ shows_nl +@+ indent s)
   })" |
  "checkTRSProof i R m (StringReversal rR prf) = debug (i []) ''String Reversal'' (do {
     check_string_rev R rR
       <+? (\<lambda>s. i +@+ '': error when applying string reversal'' +#+ shows_nl +@+ s);
     checkTRSProof (i +@+ shows_string ''.1'') rR 0 prf
       <+? (\<lambda>s. i +@+ '': error below string reversal'' +#+ shows_nl +@+ indent s)
   })"

lemma checkTRSProofSound:
  fixes prof :: "('f::show,'l::show,string)TRSProof"
  assumes check: "isOK(checkTRSProof i r m prof)"
  shows "decrSN (set r) m"
using assms
proof (induct "prof" arbitrary: i r m)
  case (DPTrans P dpprf)
  let ?il = "init_lab_rules :: (('f,'l)lab shp,'v) rules \<Rightarrow> (('f,'l)lab shp,'l,'v)trsLL"
  let ?ilr = "?il (ID_list r)"
  from DPTrans have dpsSound: "isOK(check_DP r P)" and rec: "isOK(checkDPProof (i +@+ shows_string ''.1'') (?il P) ?ilr 0 (computeRuleMap ?ilr) (computeRuleMap (reverse_trs ?ilr)) dpprf)"
    by (auto simp: Let_def)
  from rec have "finite_lab_dpp (set (?il P), set (?il (ID_list r))) 0" 
    by (rule checkDPProofSound) 
  with init_lab_sound have "finite_dpp (set P, set (ID_list r))" by auto
  with dpsSound check_DP have "SN (rstep (set r))" by simp
  thus ?case by (rule SN_to_decr_SN)
next
  case (RuleRemoval rp rR prof')
  hence ok:"isOK (checkTRSProof i r m (RuleRemoval rp rR prof'))" and IH:"\<And>r' i' m'. isOK (checkTRSProof i' r' m' prof') \<Longrightarrow> decrSN (set r') m'" by auto
  from ok have rr_ok:"isOK(create_rule_removal_proc rp r rR)" and rec_ok:"isOK(checkTRSProof (i +@+ shows_string ''.1'') rR m prof')" by auto
  from rec_ok IH have SN: "decrSN (set rR) m" by auto
  show "decrSN (set r) m"
    by (rule apply_ci_lab_tt, rule create_rule_removal_proc, rule SN, rule rr_ok)
next
  case RisEmpty
  hence "SN (rstep (set r))" by auto
  thus ?case by (rule SN_to_decr_SN)
next
  case (Unlab uR prof)
  thus ?case 
  proof (cases m, simp)
    case (Suc m')
    with Unlab have ok:"isOK (checkTRSProof i r (Suc m') (Unlab uR prof))" and IH:"\<And>r' i'. isOK (checkTRSProof i' r' m' prof) \<Longrightarrow> decrSN (set r') m'" by auto
    from ok have ul_ok:"isOK(unlab_trs_tt r uR)" and rec_ok:"isOK(checkTRSProof (i +@+ shows_string ''.1'') uR m' prof)" by auto
    from rec_ok IH have SN: "decrSN (set uR) m'" by auto
    show "decrSN (set r) m" by (rule apply_sound_lab_tt[OF unlab_trs_tt_sound], rule SN, auto simp: ul_ok Suc)
  qed
next
  case (Semlab sli L lR prof)
  hence ok:"isOK (checkTRSProof i r m (Semlab sli L lR prof))" and IH:"\<And>r' i' m'. isOK (checkTRSProof i' r' m' prof) \<Longrightarrow> decrSN (set r') m'" by auto
  from ok have l_ok:"isOK(create_sem_lab_tt sli L r lR)" and rec_ok:"isOK(checkTRSProof (i +@+ shows_string ''.1'') lR (Suc m) prof)" by auto
  from rec_ok IH have SN: "decrSN (set lR) (Suc m)" by auto
  show "decrSN (set r) m" by (rule apply_sound_lab_tt[OF create_sem_lab_tt_sound SN], simp, rule l_ok)
next
  case (Rootlab lR prof)
  hence ok:"isOK (checkTRSProof i r m (Rootlab lR prof))" and IH:"\<And>r' i' m'. isOK (checkTRSProof i' r' m' prof) \<Longrightarrow> decrSN (set r') m'" by auto
  from ok have l_ok:"isOK(rl_tt r lR)" and rec_ok:"isOK(checkTRSProof (i +@+ shows_string ''.1'') lR (Suc m) prof)" by auto
  from rec_ok IH have SN: "decrSN (set lR) (Suc m)" by auto
  show "decrSN (set r) m" by (rule apply_sound_lab_tt[OF rl_tt_sound SN], simp, rule l_ok)
next
  case (Fcc fcs rR prof)
  hence ok: "isOK (checkTRSProof i r m (Fcc fcs rR prof))" and IH: "\<And>r' i' m'. isOK(checkTRSProof i' r' m' prof) \<Longrightarrow> decrSN (set r') m'" by auto
  from ok have r_ok: "isOK(fcc_tt r fcs rR)" and rec_ok: "isOK(checkTRSProof (i +@+ shows_string ''.1'') rR 0 prof)" by auto
  from rec_ok IH have SN: "decrSN (set rR) 0" by auto
  show "decrSN (set r) m" by (rule apply_sound_lab_tt[OF fcc_tt_sound_lab SN],simp,rule r_ok)
next
  case (StringReversal rR prof)
  hence ok: "isOK (checkTRSProof i r m (StringReversal rR prof))" and IH: "\<And>r' i' m'. isOK(checkTRSProof i' r' m' prof) \<Longrightarrow> decrSN (set r') m'" by auto
  from ok have r_ok: "isOK(check_string_rev r rR)" and rec_ok: "isOK(checkTRSProof (i +@+ shows_string ''.1'') rR 0 prof)" by auto
  from rec_ok IH have SN: "decrSN (set rR) 0" by auto
  show "decrSN (set r) m" by (rule apply_sound_lab_tt[OF sound_tt_impl_to_sound_lab_tt_impl[OF string_rev_sound_tt_impl] SN], simp add: r_ok)
qed


datatype ('f,'l,'v)certificationProblem = TRSProof "('f,'v)rules" "('f,'l,'v)TRSProof" | TRSDisproof "('f,'v)rules" "('f,'v)TRSDisproof"
datatype ('f,'v)inputProblem = TRS "('f,'v)rules"

fun input :: "('f,'l,'v)certificationProblem \<Rightarrow> ('f,'v)inputProblem"
where "input (TRSProof trs _) = TRS trs"
  |   "input (TRSDisproof trs _) = TRS trs"

fun desired_property :: "('f,'l,'v)certificationProblem \<Rightarrow> bool"
where "desired_property (TRSProof trs _) = SN (rstep (set trs))"
   |  "desired_property (TRSDisproof trs _) = (\<not> (SN (rstep (set trs))))"

fun xml_tag :: "('f,'l,'v)certificationProblem \<Rightarrow> string"
where "xml_tag (TRSProof _ _) = ''<trsTerminationProof>''"
  |   "xml_tag (TRSDisproof _ _) = ''<trsNonterminationProof>''"

datatype certificationResult = Certified string | Error string

fun
  certifyCertProblem ::
    "('f::show,'l::show,string)certificationProblem \<Rightarrow> certificationResult"
where
  "certifyCertProblem (TRSProof trs prf) =
     (case checkTRSProof (shows_string ''1'') (init_lab_rules trs) 0 prf of
       Inl err \<Rightarrow> Error((''termination proof not accepted'' +#+ shows_nl +@+ err) [])
     | Inr _ \<Rightarrow> Certified ''<trsTerminationProof>'')" |
  "certifyCertProblem (TRSDisproof trs prf) =
     (case checkTRSDisproof (shows_string ''1'') trs prf of
       Inl err \<Rightarrow> Error((''non-termination proof not accepted'' +#+ shows_nl +@+ err) [])
     | Inr _ \<Rightarrow> Certified ''<trsNonterminationProof>'')"

lemma certifyCertProblemSound:
  assumes answer: "certifyCertProblem (cp :: ('f :: show,'l :: show,string)certificationProblem) = Certified info"
  shows "desired_property cp \<and> xml_tag cp = info"
proof (cases cp)
  case (TRSProof trs prof)
  let ?IL = "init_lab_rules trs :: (('f,'l)lab,string) rules"
  from TRSProof answer have ok: "isOK(checkTRSProof (shows_string ''1'') ?IL 0 prof) \<and> xml_tag cp = info" 
    by (cases "checkTRSProof (shows_string ''1'') ?IL 0 prof") auto
  have "decrSN (init_lab_trs (set trs) :: (('f,'l)lab,string)trs) 0" using checkTRSProofSound ok by force
  with TRSProof ok decr_SN_to_SN[of "set trs"] answer show ?thesis by auto
next
  case (TRSDisproof trs prof)
  with answer have "isOK(checkTRSDisproof (shows_string ''1'') trs prof) \<and> xml_tag cp = info"
    by (cases "checkTRSDisproof (shows_string ''1'') trs prof") auto
  with checkTRSDisproofSound TRSDisproof show ?thesis by auto
qed

end
