(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RootLabeling imports SemanticLabeling Name DpTransImpl begin

definition root_altering :: "('f,'v)trs \<Rightarrow> ('f,'v)trs" ("'(_')a" 1000)
where "(R)a \<equiv> {(l,r) | l r. (l,r) \<in> R \<and> root l \<noteq> root r}"

text {* The following function creates a flat context with root @{text f}
and a hole at position @{text i}. All variables are taken from the list
@{text xs}. *}
definition flat_ctxt_i :: "('f,string)term list \<Rightarrow> 'f \<Rightarrow> nat \<Rightarrow> ('f,string)ctxt"
where "flat_ctxt_i xs f i \<equiv> More f (take i xs) \<box> (drop (Suc i) xs)"

definition fresh_vars :: "nat \<Rightarrow> string set \<Rightarrow> string list"
where "fresh_vars n tabus \<equiv> fresh_strings ''x'' 0 tabus n"

definition fresh_vars_list :: "nat \<Rightarrow> string list \<Rightarrow> string list"
where "fresh_vars_list n tabus \<equiv> fresh_strings_list ''x'' 0 tabus n"

lemma fresh_vars_list_sound[simp]:
  "fresh_vars_list n tabus = fresh_vars n (set tabus)"
  unfolding fresh_vars_list_def fresh_vars_def by auto

lemma flat_ctxt_i_fresh_vars_list_conv[simp]:
  "flat_ctxt_i(map Var (fresh_vars_list n tabus)) = flat_ctxt_i(map Var (fresh_vars n (set tabus)))"
by (rule ext,rule ext,unfold flat_ctxt_i_def) simp

lemma fresh_vars_fresh:
  assumes "finite V"
  shows "set(fresh_vars n V) \<inter> V = {}"
  using assms fresh_name_gen_for_strings[of "''x''" 0]
  unfolding fresh_vars_def fresh_name_gen_def by auto

lemma fresh_vars_list_fresh: "set(fresh_vars_list n vs) \<inter> set vs = {}"
  using fresh_name_gen_for_strings_list[of "''x''" 0]
  unfolding fresh_vars_list_def fresh_name_gen_list_def by auto

lemma fresh_vars_length:
  assumes "finite V" shows "length(fresh_vars n V) = n"
  using assms fresh_name_gen_for_strings[of "''x''" 0]
  unfolding fresh_name_gen_def fresh_vars_def by simp

lemma fresh_vars_list_length: "length(fresh_vars_list n vs) = n"
  using fresh_name_gen_for_strings_list[of "''x''" 0]
  unfolding fresh_name_gen_list_def fresh_vars_list_def by simp

lemma Var_inj: "inj_on Var (set vs)" unfolding inj_on_def by auto

lemma fresh_vars_distinct: assumes "finite V" shows "distinct(fresh_vars n V)"
  using assms Var_inj[of "fresh_strings ''x'' 0 V n"]
    fresh_name_gen_for_strings[of "''x''" 0]
    distinct_map[of Var "fresh_strings ''x'' 0 V n"]
  unfolding fresh_name_gen_def fresh_vars_def by auto

lemma fresh_vars_list_distinct: "distinct(fresh_vars_list n vs)"
  using Var_inj[of "fresh_strings_list ''x'' 0 vs n"]
    fresh_name_gen_for_strings_list[of "''x''" 0]
    distinct_map[of "Var" "fresh_strings_list ''x'' 0 vs n",symmetric]
  unfolding fresh_name_gen_list_def fresh_vars_list_def by auto

definition flat_ctxts :: "string set \<Rightarrow> ('f\<times>nat)set \<Rightarrow> ('f,string)ctxt set"
where "flat_ctxts tabus fs \<equiv> (\<Union>(f,n)\<in>fs.
  {flat_ctxt_i (map Var (fresh_vars n tabus)) f i | i. i \<in> {0..<n}}
)"

definition flat_ctxts_list :: "string list \<Rightarrow> ('f\<times>nat)list \<Rightarrow> ('f,string)ctxt list"
where "flat_ctxts_list tabus fs \<equiv> concat(map (\<lambda>(f,n). (
  let xs = map Var (fresh_vars_list n tabus) in
  map (flat_ctxt_i xs f) [(0::nat)..<n]
)) fs)"

lemma flat_ctxts_list_sound[simp]:
  assumes "distinct fs"
  shows "set(flat_ctxts_list tabus fs) = flat_ctxts (set tabus) (set fs)"
using assms by (induct fs) (auto simp: flat_ctxts_list_def flat_ctxts_def,force)

definition flat_ctxt_closure :: "('f\<times>nat)set \<Rightarrow> ('f,string)trs \<Rightarrow> ('f,string)trs" ("FC")
where "FC F R \<equiv> (\<Union>(l,r)\<in>(R)a. {(C\<langle>l\<rangle>,C\<langle>r\<rangle>) | C. C \<in> flat_ctxts (vars_trs R) F}) \<union> (R - (R)a)"

definition flat_ctxt_closure_list :: "('f\<times>nat)list \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)rules"
where "flat_ctxt_closure_list fs trs \<equiv> (
  let fcs      = flat_ctxts_list (vars_trs_list trs) fs in
  let (ras,rs) = partition (\<lambda>(l,r). root l \<noteq> root r) trs in
  flat_map (\<lambda>(l,r). map (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)) fcs) ras
  @ rs
)"

lemma flat_ctxts_empty[simp]: "flat_ctxts xs {} = {}" unfolding flat_ctxts_def by simp

lemma root_altering_list[simp]: "set(fst(partition (\<lambda>(l,r). root l \<noteq> root r) trs)) = (set trs)a"
  unfolding partition_filter_conv root_altering_def by auto

lemma root_altering_list_aux[simp]: "set(snd(partition (\<lambda>(l,r). root l \<noteq> root r) trs)) = set trs - (set trs)a"
  unfolding partition_filter_conv root_altering_def by auto

lemma flat_ctxt_closure_list_sound_1[simp]: 
  fixes trs::"('f,string)rules" and fs::"('f\<times>nat)list"
  assumes "distinct fs"
  shows "set(
  flat_map (\<lambda>(l,r).
    map (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)) (flat_ctxts_list (vars_trs_list trs) fs)
  ) (fst(partition (\<lambda>(l,r). root l \<noteq> root r) trs))
) = (\<Union>(l,r)\<in>(set trs)a.
  {(C\<langle>l\<rangle>,C\<langle>r\<rangle>) | C. C \<in> flat_ctxts (vars_trs(set trs)) (set fs)}
)"
proof -
  let ?C = "\<lambda>(l,r)C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)"
  let ?ra = "fst(partition (\<lambda>(l,r). root l \<noteq> root r) trs)"
  let ?fcs = "flat_ctxts_list (vars_trs_list trs) fs"
  have "set(flat_map  (\<lambda>(l,r). map (?C(l,r)) ?fcs) ?ra) =
    (\<Union>x\<in>set(map (\<lambda>(l,r). map (?C(l,r)) ?fcs) ?ra). set x)" by simp
  also have "\<dots> = (\<Union>x\<in>(\<lambda>(l,r). map (?C(l,r)) ?fcs) ` set ?ra. set x)" by simp
  also have "\<dots> = (\<Union>(l,r)\<in>set ?ra. set((\<lambda>(l,r). map (?C(l,r)) ?fcs) (l,r)))" by best
  also have "\<dots> = (\<Union>(l,r)\<in>set ?ra. set(map (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)) ?fcs))" by simp
  also have "\<dots> = (\<Union>(l,r)\<in>set ?ra. {(C\<langle>l\<rangle>,C\<langle>r\<rangle>) | C. C \<in> set(?fcs)})" by auto
  also have "\<dots> = (\<Union>(l,r)\<in>(set trs)a. {(C\<langle>l\<rangle>,C\<langle>r\<rangle>) | C. C \<in> set(?fcs)})"
    using root_altering_list[of "trs"] by simp
  finally show ?thesis
  using flat_ctxts_list_sound[of "fs" "vars_trs_list trs",OF `distinct fs`] by auto
qed

lemma flat_ctxt_closure_list_sound[simp]:
  fixes trs::"('f,string)rules"
  assumes "distinct fs"
  shows "set(flat_ctxt_closure_list fs trs) = FC(set fs)(set trs)"
proof -
  let ?fcs = "flat_ctxts_list (vars_trs_list trs) fs"
  let ?ras = "fst(partition (\<lambda>(l,r). root l \<noteq> root r) trs)"
  let ?rs  = "snd(partition (\<lambda>(l,r). root l \<noteq> root r) trs)"
  have a: "set(
    let fcs      = ?fcs in
    let (ras,rs) = partition (\<lambda>(l,r). root l \<noteq> root r) trs in
    flat_map (\<lambda>(l,r). map (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)) fcs) ras @ rs) =
    set(flat_map (\<lambda>(l,r). map (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>)) ?fcs) ?ras) \<union> set ?rs"
    by simp
  show ?thesis
    unfolding flat_ctxt_closure_list_def flat_ctxt_closure_def
    unfolding a
    unfolding flat_ctxt_closure_list_sound_1[OF `distinct fs`]
    unfolding root_altering_list_aux by simp
qed

lemma fccE[elim]:
  fixes R::"('f,string)trs"
  assumes "(s,t) \<in> FC F R"
  and ra: "(s,t) \<in> R - (R)a \<Longrightarrow> P"
  and rp: "\<exists>(l,r)\<in>(R)a.\<exists>C\<in>flat_ctxts (vars_trs R) F. s = C\<langle>l\<rangle> \<and> t = C\<langle>r\<rangle> \<Longrightarrow> P"
  shows "P"
using assms unfolding flat_ctxts_def flat_ctxt_closure_def by force

lemma fccI_altering[intro]:
  fixes R::"('f,string)trs" assumes "C \<in> flat_ctxts (vars_trs R) F"
  and "(l,r) \<in> (R)a"
  shows "(C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> FC F R"
using assms unfolding flat_ctxt_closure_def by auto

lemma fccI_preserving[intro]:
  fixes R::"('f,string)trs" assumes "(l,r) \<in> R - (R)a" shows "(l,r) \<in> FC F R"
using assms unfolding flat_ctxt_closure_def by simp

lemma partial_fcc_preserves_SN:
  fixes R::"('f,string)trs"
  assumes SN: "SN(rstep R)"
    and refl: "\<forall>(l,r)\<in>R'.\<exists>(u,v)\<in>R. \<exists>C. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
    shows "SN(rstep R')"
using SN proof (rule contrapos_pp)
  let ?r = "rstep R"
  let ?c = "rstep R'"
  assume "\<not> SN ?c"
  then obtain t where seq: "\<forall>i. (t i,t(Suc i)) \<in> ?c" unfolding SN_defs by auto
  have "\<forall>i. (t i,t(Suc i)) \<in> ?r"
  proof
    fix i
    from seq have "(t i,t(Suc i)) \<in> ?c" by simp
    from rstep_imp_C_s_r[OF this]
      obtain C \<sigma> l r where fc: "(l,r) \<in> R'" and s: "t i = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t(Suc i) = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
    from fc refl have "\<exists>(u,v)\<in>R. \<exists>C. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>" by auto
    then obtain u v C' where R: "(u,v) \<in> R" and l: "l = C'\<langle>u\<rangle>" and r: "r = C'\<langle>v\<rangle>" by auto
    from R have "(u\<cdot>\<sigma>,v\<cdot>\<sigma>) \<in> CS R" by auto
    hence "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?r" unfolding l r rstep_CC_CS by auto
    thus "(t i,t(Suc i)) \<in> ?r" unfolding s t by auto
  qed
  thus "\<not> SN ?r" unfolding SN_defs by auto
qed

(*
lemma partial_fcc_preserves_SN:
  fixes R::"('f,string)trs"
    and F::"('f\<times>nat)set"
  defines "fcs \<equiv> flat_ctxts (vars R) F"
  assumes SN: "SN(rstep R)"
    and refl: "\<forall>(l,r)\<in>R'. (l,r) \<in> R \<or> (\<exists>(u,v)\<in>R. \<exists>C\<in>fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>)"
    shows "SN(rstep R')"
using SN proof (rule contrapos_pp)
  let ?r = "rstep R"
  let ?c = "rstep R'"
  assume "\<not> SN ?c"
  then obtain t where seq: "\<forall>i. (t i,t(Suc i)) \<in> ?c" unfolding SN_defs by auto
  have "\<forall>i. (t i,t(Suc i)) \<in> ?r"
  proof
    fix i
    from seq have "(t i,t(Suc i)) \<in> ?c" by simp
    from rstep_imp_C_s_r[OF this]
      obtain C \<sigma> l r where fc: "(l,r) \<in> R'" and s: "t i = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t(Suc i) = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
    from fc refl have "(l,r) \<in> R \<or> (\<exists>(u,v)\<in>R. \<exists>C\<in>fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>)" by auto
    thus "(t i,t(Suc i)) \<in> ?r"
    proof
      assume "(l,r) \<in> R"
      hence "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> CS R" by auto
      thus ?thesis unfolding s t by auto
    next
      assume "\<exists>(u,v)\<in>R.\<exists>C\<in>fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
      then obtain u v C' where R: "(u,v) \<in> R" and C': "C' \<in> fcs" and l: "l = C'\<langle>u\<rangle>" and r: "r = C'\<langle>v\<rangle>" by auto
      from R have "(u\<cdot>\<sigma>,v\<cdot>\<sigma>) \<in> CS R" by auto
      hence "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?r" unfolding l r ctxt_sub_distr by auto
      thus ?thesis unfolding s t by (rule rstep_imp_ctxt_rstep)
    qed
  qed
  thus "\<not> SN ?r" unfolding SN_defs by auto
qed
*)

lemma fcc_imp_refl:
  fixes R::"('f,string)trs"
    and F::"('f\<times>nat)set"
  shows "\<forall>(l,r)\<in>FC F R. (\<exists>(u,v)\<in>R. \<exists>C. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>)"
proof -
  {
    fix l r assume "(l,r) \<in> FC F R"
    hence "\<exists>(u,v)\<in>R. \<exists>C. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
    proof
      assume "(l,r) \<in> R - (R)a"
      hence "(l,r) \<in> R" by simp
      moreover have "l = \<box>\<langle>l\<rangle>" by simp
      moreover have "r = \<box>\<langle>r\<rangle>" by simp
      ultimately show ?thesis by best
    next
      assume "\<exists>(x,y)\<in>(R)a.\<exists>C\<in>flat_ctxts (vars_trs R) F. l = C\<langle>x\<rangle> \<and> r = C\<langle>y\<rangle>"
      moreover have "(R)a \<subseteq> R" unfolding root_altering_def by auto
      ultimately show ?thesis by auto
    qed
  }
  thus ?thesis by auto
qed

(*
lemma fcc_imp_refl:
  fixes R::"('f,string)trs"
    and F::"('f\<times>nat)set"
  defines "fcs \<equiv> flat_ctxts (vars R) F"
  shows "\<forall>(l,r)\<in>FC F R. (l,r) \<in> R \<or> (\<exists>(u,v)\<in>R. \<exists>C\<in>fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>)"
proof -
  {
    fix l r assume "(l,r) \<in> FC F R"
    hence "(l,r) \<in> R \<or> (\<exists>(u,v)\<in>R. \<exists>C\<in>fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>)"
    proof
      assume "(l,r) \<in> R - (R)a" thus ?thesis by auto
    next
      assume "\<exists>(x,y)\<in>(R)a.\<exists>C\<in>flat_ctxts (vars R) F. l = C\<langle>x\<rangle> \<and> r = C\<langle>y\<rangle>"
      moreover have "(R)a \<subseteq> R" unfolding root_altering_def by auto
      ultimately show ?thesis unfolding fcs_def by auto
    qed
  }
  thus ?thesis by auto
qed
*)

lemma fcc_imp_pres:
  fixes R::"('f,string)trs"
    and F::"('f\<times>nat)set"
  defines "fcs \<equiv> flat_ctxts (vars_trs R) F"
  shows "\<forall>(l,r)\<in>R. (l,r) \<in> FC F R \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> FC F R)"
proof (intro ballI2)
  fix l r assume "(l,r) \<in> R"
  hence "(l,r) \<in> (R)a \<or> (l,r) \<in> R - (R)a" by auto
  thus "(l,r) \<in> FC F R \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> FC F R)"
  proof
    assume Ra: "(l,r) \<in> (R)a"
    thus ?thesis unfolding fcs_def using fccI_altering[of _ R F l r] by auto
  next
    assume "(l,r) \<in> R - (R)a" thus ?thesis using fccI_preserving by best
  qed
qed

lemma fcc_preserves_SN:
  fixes R::"('f,string)trs" assumes "SN(rstep R)" shows "SN(rstep(FC F R))"
using partial_fcc_preserves_SN[OF assms,of "FC F R"] fcc_imp_refl[of F R] by simp

(*
lemma fcc_preserves_SN:
  fixes R::"('f,string)trs" assumes "SN(rstep R)" shows "SN(rstep(FC F R))"
using partial_fcc_preserves_SN[OF assms,of "FC F R" F] fcc_imp_refl[of F R] by simp
*)

lemma fresh_vars_sub:
  assumes "finite V" and "x \<in> set(fresh_vars n V)" shows "Var x\<cdot>(\<sigma>|_V) = Var x"
using fresh_vars_fresh[OF `finite V`] assms by (cases \<sigma>) auto

lemma flat_ctxt_sub_apply[simp]:
  assumes "finite V" and "C \<in> flat_ctxts V F" shows "C\<langle>t\<rangle>\<cdot>(\<sigma>|_V) = C\<langle>t\<cdot>(\<sigma>|_V)\<rangle>"
proof -
  from assms obtain f n i
    where i:"i<n" and fn: "(f,n)\<in>F" and C: "C = flat_ctxt_i (map Var (fresh_vars n V)) f i" (is "_ = flat_ctxt_i ?V f i")
    unfolding flat_ctxts_def by auto
  have fresh: "map (\<lambda>t. t\<cdot>(\<sigma>|_V)) ?V = ?V"
    using map_idI[of "?V" "\<lambda>t. t\<cdot>(\<sigma>|_V)"] fresh_vars_sub[OF `finite V`] by auto
  from C have C: "C = More f (take i ?V) \<box> (drop (Suc i) ?V)"
    unfolding flat_ctxt_i_def by simp
  show ?thesis unfolding C unfolding ctxt_apply.simps sub_apply_term.simps unfolding map_append
    unfolding take_map[symmetric]
    unfolding map.simps
    unfolding drop_map[symmetric]
    unfolding fresh by simp
qed

lemma flat_ctxt_has_fresh_vars:
  assumes finite: "finite V" and FC: "C \<in> flat_ctxts V F"
  shows "vars_ctxt C \<inter> V = {}"
proof -
  from FC obtain f n i where "(f,n) \<in> F" and "i < n" and "C = flat_ctxt_i (map Var (fresh_vars n V)) f i" (is "_ = flat_ctxt_i ?V _ _")
    unfolding flat_ctxts_def by auto
  hence C: "C = More f (take i ?V) \<box> (drop (Suc i) ?V)" unfolding flat_ctxt_i_def by simp
  have vC: "vars_ctxt C = (\<Union>set(map vars_term (take i ?V @ drop (Suc i) ?V)))" by (simp add: C)
  have"vars_ctxt C \<subseteq> \<Union>set(map vars_term ?V)" using fresh_vars_length[OF finite,of n] and `i < n` unfolding vC
    using set_take_subset[of i ?V] set_drop_subset[of "Suc i" ?V] by auto
  hence "{Var x | x. x \<in> vars_ctxt C} \<subseteq> set ?V" unfolding fresh_vars_def by auto
  with fresh_vars_fresh[OF finite,of n] show ?thesis by auto
qed

(* --> Position *)
lemma pos_exhaust_rev[case_names empty snoc]:
  assumes "p = \<epsilon> \<Longrightarrow> P" and "\<And>\<pi> i. p = \<pi>#>i \<Longrightarrow> P" shows "P"
using assms by (induct p rule:rev_induct) auto

lemma wf_FC: assumes "wf_trs R" shows "wf_trs(FC F R)"
unfolding wf_trs_def proof (intro allI impI)
  fix l r assume "(l,r) \<in> FC F R"
  thus "(\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l"
  proof
    assume "(l,r) \<in> R - (R)a" with assms show ?thesis unfolding wf_trs_def by auto
  next
    assume "\<exists>(l',r')\<in>(R)a. \<exists>C\<in>flat_ctxts (vars_trs R) F. l = C\<langle>l'\<rangle> \<and> r = C\<langle>r'\<rangle>"
    then obtain l' r' C where "(l',r') \<in> (R)a" and C: "C \<in> flat_ctxts (vars_trs R) F" and l: "l = C\<langle>l'\<rangle>" and r: "r = C\<langle>r'\<rangle>" by best
    hence R: "(l',r') \<in> R" unfolding root_altering_def by auto
    from C obtain f n i where "(f,n) \<in> F" and "i < n"
      and "C = flat_ctxt_i (map Var (fresh_vars n (vars_trs R))) f i" (is "C = flat_ctxt_i ?V f i") unfolding flat_ctxts_def by auto
    hence C: "C = More f (take i ?V) \<box> (drop (Suc i) ?V)" unfolding flat_ctxt_i_def by simp
    hence "l = Fun f (take i ?V@l'#drop (Suc i) ?V)" unfolding l by simp
    hence "\<exists>f ss. l = Fun f ss" by auto
    moreover have "vars_term r \<subseteq> vars_term l"
    proof -
      from R and assms have "vars_term r' \<subseteq> vars_term l'" unfolding wf_trs_def by auto
      with l r show ?thesis unfolding C by auto
    qed
    ultimately show ?thesis by simp
  qed
qed

(* --> Util *)
lemma distinct_take_drop: assumes dist: "distinct vs" and len: "i < length vs" shows "distinct(take i vs @ drop (Suc i) vs)" (is "distinct(?xs@?ys)")
proof -
  from id_take_nth_drop[OF len] have vs[symmetric]: "vs = ?xs @ vs!i # ?ys" .
  with dist have "distinct ?xs" and "distinct(vs!i#?ys)" and "set ?xs \<inter> set(vs!i#?ys) = {}" using distinct_append[of ?xs "vs!i#?ys"] by auto
  hence "distinct ?ys" and "set ?xs \<inter> set ?ys = {}" by auto
  with `distinct ?xs` show ?thesis using distinct_append[of ?xs ?ys] vs by simp
qed

(* --> Term *)
lemma sub_extend_flat_ctxt:
  assumes dist: "distinct vs"
    and len1: "length(take i (map Var vs)) = length ss1"
    and len2: "length(drop (Suc i) (map Var vs)) = length ss2"
    and i: "i < length vs"
  shows "More f (take i (map Var vs)) \<box> (drop (Suc i) (map Var vs)) \<cdot>c sub_extend \<sigma> (zip (take i vs@drop (Suc i) vs) (ss1@ss2)) = More f ss1 \<box> ss2"
proof -
  let ?V = "map Var vs"
  let ?vs1 = "take i vs" and ?vs2 = "drop (Suc i) vs"
  let ?ss1 = "take i ?V" and ?ss2 = "drop (Suc i) ?V"
  let ?\<sigma> = "sub_extend \<sigma> (zip (?vs1@?vs2) (ss1@ss2))"
  from len1 and len2 have len: "length(?vs1@?vs2) = length(ss1@ss2)" using i by simp
  from distinct_take_drop[OF dist i] have "distinct(?vs1@?vs2)" .
  from sub_extend_absorb[OF this len,of "\<sigma>"]
    have map: "map (\<lambda>t. t\<cdot>?\<sigma>) (?ss1@?ss2) = ss1@ss2" unfolding take_map drop_map map_append .
  from len1 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss1 = ss1" by auto
  moreover from len2 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss2 = ss2" by auto
  ultimately show ?thesis by simp
qed

lemma distinct_take_drop':
  assumes dist: "distinct vs" and len: "i < length vs"
  shows "distinct(take i vs @ drop i vs)" (is "distinct(?xs@?ys)")
proof -
  from append_take_drop_id have vs: "?xs @ ?ys = vs" .
  with dist have "distinct ?xs" and "distinct ?ys" and "set ?xs \<inter> set ?ys = {}"
    using distinct_append[of ?xs ?ys] by auto
  thus ?thesis using distinct_append[of ?xs ?ys] by simp
qed

lemma sub_extend_flat_ctxt'':
  assumes dist: "distinct vs"
    and len1: "length(take i (map Var vs)) = length ss1"
    and len2: "length(drop i (map Var vs)) = length ss2"
    and i: "i < length vs"
  shows "More f (take i (map Var vs)) \<box> (drop i (map Var vs)) \<cdot>c sub_extend \<sigma> (zip (take i vs@drop i vs) (ss1@ss2)) = More f ss1 \<box> ss2"
proof -
  let ?V = "map Var vs"
  let ?vs1 = "take i vs" and ?vs2 = "drop i vs"
  let ?ss1 = "take i ?V" and ?ss2 = "drop i ?V"
  let ?\<sigma> = "sub_extend \<sigma> (zip (?vs1@?vs2) (ss1@ss2))"
  from len1 and len2 have len: "length(?vs1@?vs2) = length(ss1@ss2)" using i by simp
  from distinct_take_drop'[OF dist i] have "distinct(?vs1@?vs2)" .
  from sub_extend_absorb[OF this len,of "\<sigma>"]
    have map: "map (\<lambda>t. t\<cdot>?\<sigma>) (?ss1@?ss2) = ss1@ss2" unfolding take_map drop_map map_append .
  from len1 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss1 = ss1" unfolding map_append by auto
  moreover from len2 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss2 = ss2" unfolding map_append by auto
  ultimately show ?thesis by simp
qed

lemma distinct_map_Var: assumes "distinct xs" shows "distinct(map Var xs)"
using assms by (induct xs) auto

lemma fc_flat:
  assumes C: "C \<in> flat_ctxts V F"
    and finite: "finite V"
  shows "vars_ctxt C \<inter> V = {} \<and> (\<exists>f ss1 ss2. C = More f ss1 \<box> ss2
    \<and> (f,Suc(length(ss1@ss2))) \<in> F
    \<and> (\<forall>t\<in>set(ss1@ss2). is_var t)
    \<and> distinct(ss1@ss2))"
proof -
  from C obtain g i m where "i < m" and "(g,m) \<in> F"
    and C': "C = More g (take i (map Var (fresh_vars m V))) \<box> (drop (Suc i) (map Var (fresh_vars m V)))" (is "C = More g ?ss1 \<box> ?ss2")
    unfolding flat_ctxts_def flat_ctxt_i_def by auto
  from flat_ctxt_has_fresh_vars[OF finite C] have fresh: "vars_ctxt C \<inter> V = {}" .
  from fresh_vars_length[OF finite] have len: "length(fresh_vars m V) = m" by simp
  from fresh_vars_distinct[OF finite] have dist: "distinct(fresh_vars m V)" by simp
  from length_take[of i "fresh_vars m V"] `i < m` have take: "length(?ss1) = i" unfolding len by simp
  from length_drop[of "Suc i" "fresh_vars m V"] have drop: "length(?ss2) = length(fresh_vars m V) - Suc i" by simp
  from take drop have m: "Suc(length(?ss1@?ss2)) = m" unfolding len using `i < m` by simp
  have F:"(g,Suc(length(?ss1@?ss2))) \<in> F" using `(g,m) \<in> F` unfolding m .
  have "\<forall>t\<in>set ?ss1. is_var t" unfolding take_map by auto
  moreover have "\<forall>t\<in>set ?ss2. is_var t" unfolding drop_map by auto
  ultimately have is_var: "\<forall>t\<in>set (?ss1@?ss2). is_var t" by auto
  have dist: "distinct(?ss1@?ss2)" using distinct_take_drop[OF dist,unfolded len,OF `i < m`]
    unfolding take_map drop_map map_append[symmetric] by (rule distinct_map_Var)
  with fresh C' F is_var dist show ?thesis by auto
qed

(*
lemma partial_fcc_reflects_SN:
  fixes R::"('f,string)trs"
    and F::"('f\<times>nat)set"
  defines "fcs \<equiv> flat_ctxts (vars R) F"
  assumes finite: "finite R"
    and subset: "funas R \<subseteq> F"
    and ne: "fcs \<noteq> {}"
    and pres: "\<forall>(l,r)\<in>R. (l,r) \<in> R' \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R')"
    and SN: "SN(rstep R')"
  shows "SN(rstep R)"
using SN proof (rule contrapos_pp)
  let ?r = "wfrstep R"
  let ?c = "rstep R'"
  assume "\<not> SN(rstep R)"
  hence "\<not> SN ?r" unfolding SN_wfrstep_SN_rstep_conv .
  then obtain t where seq: "\<forall>i. (t i,t(Suc i)) \<in> ?r" unfolding SN_defs by auto
  from ne obtain C where C: "C \<in> fcs" by auto
  have "\<forall>i. (C\<langle>t i\<rangle>,C\<langle>t(Suc i)\<rangle>) \<in> ?c"
  proof
    fix i
    from seq have "(t i,t(Suc i)) \<in> ?r" by simp
    hence "(t i,t(Suc i)) \<in> rstep R \<and> t i \<in> Term(funas R) \<and> t(Suc i) \<in> Term(funas R)" unfolding wfrstep_def by fast
    hence rstep: "(t i,t(Suc i)) \<in> rstep R" and wfti: "t i \<in> Term(funas R)" and wftsi: "t(Suc i) \<in> Term(funas R)" by auto
    from rstep_imp_C_s_r[OF rstep] obtain D \<sigma> l r where R: "(l,r) \<in> R"
      and ti: "t i = D\<langle>l\<cdot>\<sigma>\<rangle>" and tsi: "t(Suc i) = D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
    have vl: "vars l \<subseteq> vars_trs R" using R unfolding vars_trs_def vars_rule_def by auto
    have vr: "vars r \<subseteq> vars_trs R" using R unfolding vars_trs_def vars_rule_def by auto
    from R pres have "(l,r) \<in> R' \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R')" by auto
    thus "(C\<langle>t i\<rangle>,C\<langle>t(Suc i)\<rangle>) \<in> ?c"
    proof
      assume "(l,r) \<in> R'"
      hence "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> CS R'" by auto
      hence "(D\<langle>l\<cdot>\<sigma>\<rangle>,D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c" by auto
      from rstep_imp_ctxt_rstep[OF this] show ?thesis unfolding ti tsi by simp
    next
      assume closed: "\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R'"
      show ?thesis
      proof (cases D rule: ctxt_exhaust_rev)
        case Hole
        let ?\<sigma> = "\<sigma>|_(vars_trs R)"
        from C and closed have "(C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R'" by auto
        from csI[OF this] have "(C\<langle>l\<rangle>\<cdot>?\<sigma>,C\<langle>r\<rangle>\<cdot>?\<sigma>) \<in> CS R'" by simp
        from ccI[OF this,of \<box>] have "(C\<langle>l\<rangle>\<cdot>?\<sigma>,C\<langle>r\<rangle>\<cdot>?\<sigma>) \<in> ?c" by simp
        hence "(C \<cdot>c ?\<sigma> \<langle>l\<cdot>?\<sigma>\<rangle>,C \<cdot>c ?\<sigma> \<langle>r\<cdot>?\<sigma>\<rangle>) \<in> ?c" by simp
        hence "(C \<cdot>c ?\<sigma> \<langle>l\<cdot>\<sigma>\<rangle>,C \<cdot>c ?\<sigma> \<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c"
          unfolding coincidence_lemma'[OF vl,symmetric] coincidence_lemma'[OF vr,symmetric] coincidence_lemma[symmetric] .
        thus ?thesis unfolding sub_apply_ctxt_id[OF flat_ctxt_has_fresh_vars,OF `finite R` C[unfolded fcs_def]]
          unfolding ti tsi Hole by simp
      next
        case (More E f ss1 ss2)
        from ctxt_apply_wf_imp_ctxt_and_term_wf[OF wfti[unfolded ti]]
          have D: "D \<in> Ctxt(funas R)" and "l\<cdot>\<sigma> \<in> Term(funas R)" by auto
        from ctxt_compose_wf_imp_ctxts_wf[OF D[unfolded More]] have "E \<in> Ctxt(funas R) \<and> More f ss1 \<box> ss2 \<in> Ctxt(funas R)" .
        hence "More f ss1 \<box> ss2 \<in> Ctxt(funas R)" by simp
        hence "(f,Suc(length(ss1@ss2))) \<in> funas R" (is "(f,?n) \<in> funas R") by (cases rule: wf_fun_ctxts.cases) auto
        with subset have F: "(f,?n) \<in> F" by auto
        let ?V = "fresh_vars ?n (vars R)"
        have len: "length ss1 < ?n" (is "?i < _") by simp
        with F have C': "More f (take ?i (map Var ?V)) \<box> (drop (Suc ?i) (map Var ?V)) \<in> flat_ctxts (vars R) F" (is "?C' \<in> flat_ctxts (vars R) F")
          unfolding flat_ctxts_def flat_ctxt_i_def by force
        with closed have FC_step: "(?C'\<langle>l\<rangle>,?C'\<langle>r\<rangle>) \<in> R'" unfolding fcs_def by best
        let ?\<sigma> = "sub_extend \<sigma> (zip (take ?i ?V @ drop (Suc ?i) ?V) (ss1@ss2))"
        from finite have "finite(vars R)" unfolding vars_trs_def vars_rule_def by simp
        from fresh_vars_fresh[OF this] have "vars R \<inter> set ?V = {}" by auto
        moreover have "set(take (length ss1) ?V) \<subseteq> set ?V" by (rule set_take_subset)
        moreover have "set(drop (Suc(length ss1)) ?V) \<subseteq> set ?V" by (rule set_drop_subset)
        ultimately have fresh: "vars R \<inter> set(take (length ss1) ?V @ drop (Suc(length ss1)) ?V) = {}" by auto
        from fresh_vars_distinct[OF `finite(vars R)`] have dist: "distinct ?V" .
        from len have len1: "length(take (length ss1) ?V) = length ss1" using fresh_vars_length[OF `finite(vars R)`,of ?n] by simp
        hence len2: "length(drop (Suc(length ss1)) ?V) = length ss2" using fresh_vars_length[OF `finite(vars R)`,of ?n] by auto
        from len1 have len1': "length(take (length ss1) (map Var ?V)) = length ss1" by simp
        from len2 have len2': "length(drop (Suc(length ss1)) (map Var ?V)) = length ss2" by simp
        have len3: "length ss1 < length ?V" using fresh_vars_length[OF `finite(vars R)`,of ?n] by simp
        have Cl: "(?C'\<langle>l\<rangle>)\<cdot>?\<sigma> = (More f ss1 \<box> ss2)\<langle>l\<cdot>\<sigma>\<rangle>"
          unfolding ctxt_sub_distr sub_extend_id[OF fresh vl,of \<sigma>] sub_extend_flat_ctxt[OF dist len1' len2' len3] by simp
        have Cr: "(?C'\<langle>r\<rangle>)\<cdot>?\<sigma> = (More f ss1 \<box> ss2)\<langle>r\<cdot>\<sigma>\<rangle>"
          unfolding ctxt_sub_distr sub_extend_id[OF fresh vr,of \<sigma>] sub_extend_flat_ctxt[OF dist len1' len2' len3] by simp
        let ?C'' = "More f ss1 \<box> ss2"
        from csI[OF FC_step] have "(?C''\<langle>l\<cdot>\<sigma>\<rangle>,?C''\<langle>r\<cdot>\<sigma>\<rangle>) \<in> CS R'" unfolding Cl[symmetric] Cr[symmetric] by simp
        from ccI[OF this,of \<box>] have "(?C''\<langle>l\<cdot>\<sigma>\<rangle>,?C''\<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c" by simp
        from ccI[OF this,of E] have "(E\<langle>?C''\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>,E\<langle>?C''\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>) \<in> ?c" by simp
        from ccI[OF this,of C] have "(C\<langle>E\<langle>?C''\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>\<rangle>,C\<langle>E\<langle>?C''\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>\<rangle>) \<in> ?c" by simp
        thus ?thesis unfolding ti tsi More ctxt_compose_apply_eq .
      qed
    qed
  qed
  thus "\<not> SN ?c" unfolding SN_defs by auto
qed
*)

fun is_flat_ctxt :: "'v set \<Rightarrow> ('f\<times>nat)set \<Rightarrow> ('f,'v)ctxt \<Rightarrow> bool"
where "is_flat_ctxt V F (More f ss1 \<box> ss2) = (
          (f,Suc(length(ss1@ss2))) \<in> F
        \<and> forall is_var (ss1@ss2)
        \<and> distinct(ss1@ss2)
        \<and> vars_ctxt(More f ss1 \<box> ss2) \<inter> V = {}
      )"
    | "is_flat_ctxt V F _ = False"

lemma is_flat_ctxt_cases[consumes 1,case_names True]:
  assumes "is_flat_ctxt V F C"
    and "\<And>f ss1 ss2. C = More f ss1 \<box> ss2 \<Longrightarrow> P"
  shows "P"
using assms(1) proof (cases rule: is_flat_ctxt.cases[of "(V,F,C)"])
  case (1 V' F' f ss1 ss2)
  hence "C = More f ss1 \<box> ss2" by simp
  with assms(2) show ?thesis by auto
next
  case ("2_1" V' F') with assms(1) show ?thesis by simp
next
  case ("2_2" V' F' f ss1 g ts1 D ts2 ss2) with assms(1) show ?thesis by simp
qed

lemma is_flat_ctxt_imp_flat_ctxt: 
  assumes "is_flat_ctxt V F C"
  shows "vars_ctxt C \<inter> V = {} \<and> (\<exists>f ss1 ss2. C = More f ss1 \<box> ss2
          \<and> (f,Suc(length(ss1@ss2))) \<in> F
          \<and> forall is_var (ss1@ss2)
          \<and> distinct(ss1@ss2))"
using assms by (cases rule: is_flat_ctxt_cases,auto)

lemma terms_to_vars: assumes "forall is_var ts" shows "\<Union>set(map vars_term ts) = set(map the_var ts)"
using assms proof (induct ts)
  case Nil thus ?case by simp
next
  case (Cons t ts)
  hence IH: "\<Union>set(map vars_term ts) = set(map the_var ts)" by simp
  from Cons have "is_var t" by simp
  hence "vars_term t = {the_var t}" by (induct t) auto
  with IH show ?case by simp
qed

lemma Var_the_var_id: assumes "forall is_var ts" shows "map Var (map the_var ts) = ts"
using assms proof (induct ts)
  case Nil show ?case by simp
next
  case (Cons t ts)
  hence IH: "map Var (map the_var ts) = ts" and "is_var t" by auto
  from `is_var t` have "Var(the_var t) = t" by (induct t) auto
  with IH show ?case by simp
qed

lemma distinct_the_vars:
  assumes "forall is_var ts" and "distinct ts" shows "distinct(map the_var ts)"
using assms proof (induct ts)
  case Nil show ?case by simp
next
  case (Cons t ts)
  hence IH: "distinct(map the_var ts)"
    and "distinct ts" and "forall is_var ts" and "t \<notin> set ts" and "is_var t" by auto
  have "the_var t \<notin> set(map the_var ts)"
  proof
    assume "the_var t \<in> set(map the_var ts)"
    hence "Var(the_var t) \<in> set(map Var (map the_var ts))" unfolding map_map by auto
    moreover have "set(map Var (map the_var ts)) = set ts" using Var_the_var_id[OF `forall is_var ts`] by simp
    ultimately have "Var(the_var t) \<in> set ts" by best
    with `is_var t` have "t \<in> set ts" by (induct t) auto
    with `t \<notin> set ts` show False by simp
  qed
  with IH show ?case by simp
qed

lemma sub_extend_flat_ctxt':
  assumes dist: "distinct(vs1@vs2)"
    and len1: "length vs1 = length ss1"
    and len2: "length vs2 = length ss2"
  shows "More f (map Var vs1) \<box> (map Var vs2) \<cdot>c sub_extend \<sigma> (zip (vs1@vs2) (ss1@ss2)) = More f ss1 \<box> ss2"
proof -
  let ?V = "map Var (vs1@vs2)"
  let ?vs1 = "vs1" and ?vs2 = "vs2"
  let ?ss1 = "map Var vs1" and ?ss2 = "map Var vs2"
  let ?\<sigma> = "sub_extend \<sigma> (zip (?vs1@?vs2) (ss1@ss2))"
  from len1 and len2 have len: "length(?vs1@?vs2) = length(ss1@ss2)" by simp
  from sub_extend_absorb[OF dist len,of "\<sigma>"] have map: "map (\<lambda>t. t\<cdot>?\<sigma>) (?ss1@?ss2) = ss1@ss2" unfolding map_append .
  from len1 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss1 = ss1" by auto
  moreover from len2 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss2 = ss2" by auto
  ultimately show ?thesis by simp
qed

lemma partial_fcc_reflects_SN:
  fixes R::"('f,string)trs"
  assumes flat_ctxts: "\<forall>C\<in>fcs. is_flat_ctxt (vars_trs R) F C"
    and complete: "\<forall>(f,n)\<in>F.\<forall>i<n.\<exists>ss1 ss2. (More f ss1 \<box> ss2) \<in> fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
    and subset: "funas_trs R \<subseteq> F"
    and ne: "fcs \<noteq> {}"
    and pres: "\<forall>(l,r)\<in>R. (l,r) \<in> R' \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R')"
    and SN: "SN(rstep R')"
  shows "SN(rstep R)"
using SN proof (rule contrapos_pp)
  let ?r = "wfrstep R"
  let ?c = "rstep R'"
  assume "\<not> SN(rstep R)"
  hence "\<not> SN ?r" unfolding SN_wfrstep_SN_rstep_conv .
  then obtain t where seq: "\<forall>i. (t i,t(Suc i)) \<in> ?r" unfolding SN_defs by auto
  from ne obtain C where C: "C \<in> fcs" by auto
  from flat_ctxts[THEN bspec[where x=C],OF C] have distinct: "vars_ctxt C \<inter> vars_trs R = {}"
    using is_flat_ctxt_imp_flat_ctxt by best
  have "\<forall>i. (C\<langle>t i\<rangle>,C\<langle>t(Suc i)\<rangle>) \<in> ?c"
  proof
    fix i
    from seq have "(t i,t(Suc i)) \<in> ?r" by simp
    hence "(t i,t(Suc i)) \<in> rstep R \<and> t i \<in> wf_terms (funas_trs R)
      \<and> t(Suc i) \<in> wf_terms (funas_trs R)" unfolding wfrstep_def by fast
    hence rstep: "(t i,t(Suc i)) \<in> rstep R" and wfti: "t i \<in> wf_terms (funas_trs R)"
      and wftsi: "t(Suc i) \<in> wf_terms (funas_trs R)" by auto
    from rstep_imp_C_s_r[OF rstep] obtain D \<sigma> l r where R: "(l,r) \<in> R"
      and ti: "t i = D\<langle>l\<cdot>\<sigma>\<rangle>" and tsi: "t(Suc i) = D\<langle>r\<cdot>\<sigma>\<rangle>" by auto
    have vl: "vars_term l \<subseteq> vars_trs R" using R unfolding vars_trs_def vars_rule_def by auto
    have vr: "vars_term r \<subseteq> vars_trs R" using R unfolding vars_trs_def vars_rule_def by auto
    from R pres have "(l,r) \<in> R' \<or> (\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R')" by auto
    thus "(C\<langle>t i\<rangle>,C\<langle>t(Suc i)\<rangle>) \<in> ?c"
    proof
      assume "(l,r) \<in> R'"
      hence "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> CS R'" by auto
      hence "(D\<langle>l\<cdot>\<sigma>\<rangle>,D\<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c" unfolding rstep_CC_CS by auto
      from rstep.ctxt[OF this] show ?thesis unfolding ti tsi by simp
    next
      assume closed: "\<forall>C\<in>fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R'"
      show ?thesis
      proof (cases D rule: ctxt_exhaust_rev)
        case Hole
        let ?\<sigma> = "\<sigma>|_(vars_trs R)"
        from C and closed have "(C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> R'" by auto
        from sub_closureI[OF this] have "(C\<langle>l\<rangle>\<cdot>?\<sigma>,C\<langle>r\<rangle>\<cdot>?\<sigma>) \<in> CS R'" by simp
        from ctxt_closureI[OF this,of \<box>] have "(C\<langle>l\<rangle>\<cdot>?\<sigma>,C\<langle>r\<rangle>\<cdot>?\<sigma>) \<in> ?c" unfolding rstep_CC_CS
          by simp
        hence "((C \<cdot>c ?\<sigma>)\<langle>l\<cdot>?\<sigma>\<rangle>, (C \<cdot>c ?\<sigma>)\<langle>r\<cdot>?\<sigma>\<rangle>) \<in> ?c" by simp
        hence "((C \<cdot>c ?\<sigma>)\<langle>l\<cdot>\<sigma>\<rangle>, (C \<cdot>c ?\<sigma>)\<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c"
          unfolding coincidence_lemma'[OF vl,symmetric] coincidence_lemma'[OF vr,symmetric] coincidence_lemma[symmetric] .
        thus ?thesis unfolding sub_apply_ctxt_id[OF distinct] unfolding ti tsi Hole by simp
      next
        case (More E f ss1 ss2)
        from ctxt_apply_wf_imp_ctxt_and_term_wf[OF wfti[unfolded ti]]
          have D: "D \<in> wf_ctxts (funas_trs R)" and "l\<cdot>\<sigma> \<in> wf_terms (funas_trs R)" by auto
        from ctxt_compose_wf_imp_ctxts_wf[OF D[unfolded More]]
          have "E \<in> wf_ctxts (funas_trs R) \<and> More f ss1 \<box> ss2 \<in> wf_ctxts (funas_trs R)" .
        hence "More f ss1 \<box> ss2 \<in> wf_ctxts (funas_trs R)" by simp
        hence "(f,Suc(length(ss1@ss2))) \<in> funas_trs R" (is "(f,?n) \<in> funas_trs R")
          by (cases rule: wf_ctxts.cases) auto
        with subset have F: "(f,?n) \<in> F" by auto
        have len: "length ss1 < ?n" (is "?i < _") by simp
        from complete[THEN bspec[where x="(f,?n)"],OF F] len
          obtain  ts1 ts2 where C': "More f ts1 \<box> ts2 \<in> fcs" (is "?C' \<in> fcs")
          and "length ts1 = ?i" and "length ts2 = ?n - ?i - 1" by best
        with closed have FC_step: "(?C'\<langle>l\<rangle>,?C'\<langle>r\<rangle>) \<in> R'" by best
        have len_ss1: "length ss1 = length ts1" using `length ts1 = ?i` by simp
        have len_ss2: "length ss2 = length ts2" using `length ts2 = ?n - ?i - 1` by simp
        have len_ss1ss2: "length(ss1@ss2) = length(ts1@ts2)" using len_ss1 len_ss2 by simp
        let ?V = "map the_var (ts1@ts2)"
        let ?\<sigma> = "sub_extend \<sigma> (zip ?V (ss1@ss2))"
        from is_flat_ctxt_imp_flat_ctxt[OF flat_ctxts[THEN bspec,OF C']]
          have is_var: "forall is_var (ts1@ts2)" and fresh: "vars_ctxt ?C' \<inter> vars_trs R = {}" by auto
        have vars_C': "vars_ctxt ?C' = set ?V" using terms_to_vars[OF is_var] by simp
        from fresh have fresh: "vars_trs R \<inter> set ?V = {}" unfolding vars_C' by auto
        from is_flat_ctxt_imp_flat_ctxt[OF flat_ctxts[THEN bspec,OF C']]
          have "forall is_var (ts1@ts2)" and "distinct(ts1@ts2)" by auto
        hence "distinct ?V" by (rule distinct_the_vars)
        hence dist: "distinct(map the_var ts1 @ map the_var ts2)" by simp
        have len1: "length(map the_var ts1) = length ss1" unfolding len_ss1 by simp
        have len2: "length(map the_var ts2) = length ss2" unfolding len_ss2 by simp
        from is_var have "forall is_var ts1" by simp
        hence id1: "map Var (map the_var ts1) = ts1" by (rule Var_the_var_id)
        from is_var have "forall is_var ts2" by simp
        hence id2: "map Var (map the_var ts2) = ts2" by (rule Var_the_var_id)
        have Cl: "(?C'\<langle>l\<rangle>)\<cdot>?\<sigma> = (More f ss1 \<box> ss2)\<langle>l\<cdot>\<sigma>\<rangle>"
          unfolding sub_apply_term_ctxt_apply_distrib sub_extend_id[OF fresh vl,of \<sigma>]
          unfolding map_append sub_extend_flat_ctxt'[OF dist len1 len2,unfolded id1 id2] by simp
        have Cr: "(?C'\<langle>r\<rangle>)\<cdot>?\<sigma> = (More f ss1 \<box> ss2)\<langle>r\<cdot>\<sigma>\<rangle>"
          unfolding sub_apply_term_ctxt_apply_distrib sub_extend_id[OF fresh vr,of \<sigma>]
          unfolding map_append sub_extend_flat_ctxt'[OF dist len1 len2,unfolded id1 id2] by simp
        let ?C'' = "More f ss1 \<box> ss2"
        from sub_closureI[OF FC_step] have "(?C''\<langle>l\<cdot>\<sigma>\<rangle>,?C''\<langle>r\<cdot>\<sigma>\<rangle>) \<in> CS R'" unfolding Cl[symmetric] Cr[symmetric] by simp
        from ctxt_closureI[OF this,of \<box>] have "(?C''\<langle>l\<cdot>\<sigma>\<rangle>,?C''\<langle>r\<cdot>\<sigma>\<rangle>) \<in> ?c"
          unfolding rstep_CC_CS by simp
        from ctxt_closureI[OF this,of E] have "(E\<langle>?C''\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>,E\<langle>?C''\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>) \<in> ?c" by simp
        from ctxt_closureI[OF this,of C] have "(C\<langle>E\<langle>?C''\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>\<rangle>,C\<langle>E\<langle>?C''\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>\<rangle>) \<in> ?c" by simp
        thus ?thesis unfolding ti tsi More by simp
      qed
    qed
  qed
  thus "\<not> SN ?c" unfolding SN_defs by auto
qed

lemma flat_ctxt_is_flat: assumes fin: "finite V" and fc: "C \<in> flat_ctxts V F" shows "is_flat_ctxt V F C"
using fc_flat[OF fc fin] by auto

lemma fcc_reflects_SN:
  fixes R::"('f,string)trs"
  assumes finite: "finite R"
    and subset: "funas_trs R \<subseteq> G"
    and ne: "flat_ctxts (vars_trs R) G \<noteq> {}" (is "?fcs \<noteq> {}")
    and SN: "SN(rstep(FC G R))"
  shows "SN(rstep R)"
proof -
  from finite have fin: "finite(vars_trs R)" unfolding vars_trs_def vars_rule_def by auto
  have fcs: "\<forall>C\<in>?fcs. is_flat_ctxt (vars_trs R) G C" using flat_ctxt_is_flat[OF fin] by best
  have complete: "\<forall>(f,n)\<in>G.\<forall>i<n.\<exists>ss1 ss2. More f ss1 \<box> ss2 \<in> ?fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
  proof (intro ballI2 allI impI)
    fix f n i
    let ?V = "map Var (fresh_vars n (vars_trs R))"
    assume "(f,n) \<in> G" and "i < n"
    hence "flat_ctxt_i ?V f i \<in> ?fcs" unfolding flat_ctxts_def by auto
    hence "More f (take i ?V) \<box> (drop (Suc i) ?V) \<in> ?fcs" (is "More f ?ss1 \<box> ?ss2 \<in> _") unfolding flat_ctxt_i_def by auto
    moreover have "length ?ss1 = i" using fresh_vars_length[OF fin,of n] `i < n` by simp
    moreover have "length ?ss2 = n - i - 1" using fresh_vars_length[OF fin,of n] `i < n` by simp
    ultimately show "\<exists>ss1 ss2. More f ss1 \<box> ss2 \<in> ?fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1" by best
  qed
  from partial_fcc_reflects_SN[OF fcs complete subset ne fcc_imp_pres SN] show ?thesis .
qed

lemma SN_iff_fcc_SN:
  fixes R::"('f,string)trs"
  assumes "finite R" and "funas_trs R \<subseteq> G" and "flat_ctxts (vars_trs R) G \<noteq> {}"
  shows "(SN(rstep R)::bool) = SN(rstep(FC G R))"
using fcc_reflects_SN[OF assms] fcc_preserves_SN[of R G] by best

lemma fcc_ne_imp_funs_ne:
  fixes R::"('f,string)trs"
  assumes ne: "flat_ctxts V (funas_trs R) \<noteq> {}"
  shows "funs_trs R \<noteq> {}"
proof -
  from ne have "funas_trs R \<noteq> {}" unfolding flat_ctxts_def by auto
  then obtain l r where R: "(l,r) \<in> R" and "funas_term l \<union> funas_term r \<noteq> {}" unfolding funas_rule_def funas_trs_def by force
  hence "funas_term l \<noteq> {} \<or> funas_term r \<noteq> {}" by simp
  thus "funs_trs R \<noteq> {}"
  proof
    assume "funas_term l \<noteq> {}"
    then obtain f ss where "l = Fun f ss" by (induct l) auto
    hence "f \<in> funs_term l" by simp
    with R show ?thesis unfolding funs_trs_def funs_rule_def by force
  next
    assume "funas_term r \<noteq> {}"
    then obtain f ss where "r = Fun f ss" by (induct r) auto
    hence "f \<in> funs_term r" by simp
    with R show ?thesis unfolding funs_trs_def funs_rule_def by force
  qed
qed

(* Closure under Flat Contexts *)
definition
  check_rule_preserving ::
    "('f::show, string) ctxt list
     \<Rightarrow> ('f, string) rule list
     \<Rightarrow> ('f, string) rule
     \<Rightarrow> shows check"
where
  "check_rule_preserving fcs rs' rule \<equiv> (
     check (rule mem rs' \<or> forall (\<lambda>C. (C\<langle>fst rule\<rangle>,C\<langle>snd rule\<rangle>) mem rs') fcs)
       (''the rule '' +#+ shows_rule rule +@+
         '' is neither contained in the resulting set of rules nor closed under all flat contexts''
         +#+ shows_nl)
   )"

definition
  check_rule_reflecting ::
    "('f::show, string) ctxt list
     \<Rightarrow> ('f, string) rule list
     \<Rightarrow> ('f, string) rule
     \<Rightarrow> shows check"
where
  "check_rule_reflecting fcs rs rule \<equiv> (
     check (any (\<lambda>(l,r). any (\<lambda>C. fst rule = C\<langle>l\<rangle> \<and> snd rule = C\<langle>r\<rangle>) (\<box>#fcs)) rs)
       (''the rule '' +#+ shows_rule rule +@+
         '' is neither contained in the original set of rules nor obtained by applying a flat context''
         +#+ shows_nl)
   )"

lemma check_rule_preserving_sound:
  assumes "isOK(check_rule_preserving fcs rs' (l,r))"
  shows "(l,r) \<in> set rs' \<or> (\<forall>C\<in>set fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> set rs')"
proof -
  from assms have "(l,r) mem rs' \<or> forall (\<lambda>C. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) mem rs') fcs" unfolding check_rule_preserving_def by simp
  thus ?thesis unfolding mem_iff by simp
qed

lemma check_rule_reflecting_sound:
  assumes "isOK(check_rule_reflecting fcs rs (l,r))"
  shows "\<exists>(u,v) \<in> set rs.\<exists>C\<in>set(\<box>#fcs). l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
proof -
  from assms have "any (\<lambda>(u,v). any (\<lambda>C. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>) (\<box>#fcs)) rs" unfolding check_rule_reflecting_def by simp
  thus ?thesis unfolding any_set_conv .
qed

lemma check_rules_preserving_sound:
  assumes "isOK(check_allm (check_rule_preserving fcs lR) R)"
  shows "\<forall>(l,r)\<in>set R. (l,r) \<in> set lR \<or> (\<forall>C\<in>set fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> set lR)"
proof -
  from assms have "\<forall>rule\<in>set R. isOK(check_rule_preserving fcs lR rule)" by simp
  thus ?thesis using check_rule_preserving_sound by best
qed

lemma check_rules_reflecting_sound:
  assumes "isOK(check_allm (check_rule_reflecting fcs R) lR)"
  shows "\<forall>(l,r)\<in>set lR. \<exists>(u,v)\<in>set R. \<exists>C\<in>set(\<box>#fcs). l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
proof -
  from assms have "\<forall>rule\<in>set lR. isOK(check_rule_reflecting fcs R rule)" by simp
  thus ?thesis using check_rule_reflecting_sound[of fcs R] by fast
qed

definition
  check_flat_ctxt :: "string list \<Rightarrow> ('f::show, string) ctxt \<Rightarrow> shows check"
where
  "check_flat_ctxt vs C \<equiv>
     (case C of
       More f ss1 \<box> ss2 \<Rightarrow> do {
         check (distinct (ss1 @ ss2)) (shows C +@+ '' contains duplicate variables'' +#+ shows_nl);
         check (forall is_var (ss1@ss2)) (shows C
           +@+ '' is not flat, i.e., has depth greater than one'' +#+ shows_nl);
         check (forall (\<lambda>t. \<not> (the_var t) mem vs) (ss1 @ ss2)) (shows C +@+
           '' has to contain only fresh variables'' +#+ shows_nl)
       }
     | _ \<Rightarrow> error (shows C +@+ '' is not a flat context'' +#+ shows_nl))"

fun
  is_flat_ctxt_list :: "'v list \<Rightarrow> ('f\<times>nat)list \<Rightarrow> ('f,'v)ctxt \<Rightarrow> bool"
where
  "is_flat_ctxt_list vs fas (More f ss1 \<box> ss2) = (
     (f,Suc(length(ss1@ss2))) mem fas
       \<and> forall is_var (ss1@ss2) \<and> distinct(ss1@ss2) \<and> list_inter (map the_var (ss1@ss2)) vs = []
   )" |
  "is_flat_ctxt_list vs fas _ = False"

lemma is_flat_ctxt_list_cases[consumes 1,case_names True]:
  assumes "is_flat_ctxt_list vs fas C" and "\<And>f ss1 ss2. C = More f ss1 \<box> ss2 \<Longrightarrow> P" shows "P"
using assms by (cases rule: is_flat_ctxt_list.cases[of "(vs,fas,C)"]) auto

lemma is_flat_ctxt_list_sound: "is_flat_ctxt_list vs fas C = is_flat_ctxt (set vs) (set fas) C"
proof
  assume fc: "is_flat_ctxt_list vs fas C" thus "is_flat_ctxt (set vs) (set fas) C"
  proof (cases rule: is_flat_ctxt_list_cases)
    case (True f ss1 ss2)
    from fc have "(f,Suc(length(ss1@ss2))) \<in> set fas"
      and "forall is_var (ss1@ss2)" and "distinct(ss1@ss2)"
      and "set(map the_var (ss1@ss2)) \<inter> set vs = {}"
      unfolding True  mem_iff[symmetric] list_inter_conv[symmetric] by auto
    moreover have "set(map the_var (ss1@ss2)) = vars_ctxt C" unfolding True
      using terms_to_vars[OF `forall is_var (ss1@ss2)`] by simp
    ultimately show ?thesis unfolding True by simp
  qed
next
  assume fc: "is_flat_ctxt (set vs) (set fas) C" thus "is_flat_ctxt_list vs fas C"
  proof (cases rule: is_flat_ctxt_cases)
    case (True f ss1 ss2)
       from fc have "(f,Suc(length(ss1@ss2))) mem fas"
      and "forall is_var (ss1@ss2)" and "distinct(ss1@ss2)"
      and "list_inter (map the_var (ss1@ss2)) vs = []"
      unfolding True mem_iff using terms_to_vars[of "ss1@ss2"] list_inter_conv[of "map the_var (ss1@ss2)" "vs"] by auto
    thus ?thesis unfolding True by simp
  qed
qed

fun hole_at where
  "hole_at n i f (More g ss1 \<box> ss2) = (g = f \<and> length ss1 = i \<and> length ss2 = n - i - 1)" |
  "hole_at n i f _ = False"

definition
  check_flat_ctxt_complete :: "('f::show,string)ctxt list \<Rightarrow> ('f\<times>nat) \<Rightarrow> shows check"
where
  "check_flat_ctxt_complete fcs fa \<equiv>
     check (forall (\<lambda>i. any (hole_at (snd fa) i (fst fa)) fcs) [0..<snd fa])
       (''the list of flat contexts is incomplete'' +#+ shows_nl)"

definition
  check_is_flat_ctxt :: "string list \<Rightarrow> ('f::show\<times>nat)list \<Rightarrow> ('f,string)ctxt \<Rightarrow> shows check"
where
  "check_is_flat_ctxt vs fas C \<equiv> (
     check (is_flat_ctxt_list vs fas C) (shows C +@+ '' is not a flat context'' +#+ shows_nl))"

definition fcc_tt ::
  "('f::show, string) rules
   \<Rightarrow> ('f, string) ctxt list
   \<Rightarrow> ('f, string) rules
   \<Rightarrow> shows check"
where
  "fcc_tt R fcs lR \<equiv> do {
     let vs = vars_trs_list R;
     let fas = funas_trs_list R;
     check (fcs \<noteq> []) ('' at least one flat context is required for the closure under flat contexts''
       +#+ shows_nl);
     check_allm (check_flat_ctxt vs) fcs;
     check_allm (check_is_flat_ctxt vs fas) fcs;
     check_allm (check_flat_ctxt_complete fcs) fas;
     check_allm (check_rule_preserving fcs lR) R;
     check_allm (check_rule_reflecting fcs R) lR
   }"

lemma check_flat_ctxt_complete_sound:
  assumes "isOK (check_allm (check_flat_ctxt_complete fcs) F)"
  shows "\<forall>(f, n)\<in>set F. \<forall>i<n. \<exists>ss1 ss2.
    More f ss1 \<box> ss2 \<in> set fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
proof (intro ballI2 allI impI)
  fix f n i assume "(f, n) \<in> set F" and "i < n"
  with assms have "any (hole_at n i f) fcs"
    using check_flat_ctxt_complete_def[of fcs "(f, n)"] by auto
  then obtain C where "C \<in> set fcs" and C: "hole_at n i f C" unfolding any_set_conv by auto
  from C show "\<exists>ss1 ss2. More f ss1 \<box> ss2 \<in> set fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
  proof (cases rule: hole_at.cases[of "(n, i, f, C)"])
    case (1 n' i' f' g ss1 ss2) with `C \<in> set fcs` and C show ?thesis by auto
  next
    case ("2_1" n' i' f') with C show ?thesis by simp
  next
    case ("2_2" n' i' f' g ss1 h ts1 D ts2 ss2) with C show ?thesis by simp
  qed
qed

lemma fcc_tt_sound: "sound_tt_impl(\<lambda>R rR. isOK(fcc_tt R fcs rR))"
unfolding sound_tt_impl_def
proof (intro allI impI)
  fix R rR assume ok: "isOK(fcc_tt R fcs rR)" and SN: "SN(rstep(set rR))"
  from ok[unfolded fcc_tt_def Let_def] have "\<forall>C\<in>set fcs. is_flat_ctxt (vars_trs(set R)) (funas_trs(set R)) C"
    using is_flat_ctxt_list_sound[of "vars_trs_list R" "funas_trs_list R"]
    check_is_flat_ctxt_def[of "vars_trs_list R" "funas_trs_list R"] by auto
  moreover have "funas(set R) \<subseteq> funas(set R)" by simp
  moreover
    have "\<forall>(f,n)\<in>funas_trs(set R).\<forall>i<n.\<exists>ss1 ss2.
    More f ss1 \<box> ss2 \<in> set fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
    using ok[unfolded fcc_tt_def Let_def]
    check_flat_ctxt_complete_sound[of fcs "(funas_trs_list R)"] by auto
  moreover from ok[unfolded fcc_tt_def Let_def] have "fcs \<noteq> []" by simp
  moreover from ok[unfolded fcc_tt_def Let_def] check_rules_preserving_sound[of fcs rR R]
  have "\<forall>(l,r)\<in>set R. (l,r) \<in> set rR \<or> (\<forall>C\<in>set fcs. (C\<langle>l\<rangle>,C\<langle>r\<rangle>) \<in> set rR)" by auto
  ultimately show "SN(rstep(set R))"
    using SN partial_fcc_reflects_SN[of "set fcs" "set R" "funas_trs(set R)" "set rR"] by simp
qed

lemma fcc_tt_sound_lab: "sound_lab_tt_impl(\<lambda>R m lR rm. rm = 0 \<and> isOK(fcc_tt R fcs lR))"
  using sound_tt_impl_to_sound_lab_tt_impl[OF fcc_tt_sound] .

fun
  block_term :: "'f \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term"
where
  "block_term f (Var x) = Var x" |
  "block_term f (Fun g ts) = Fun g (map (\<lambda>t. Fun f [t]) ts)"

definition
  block_rule :: "'f \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) rule"
where
  "block_rule f r = (block_term f (fst r), block_term f (snd r))"

definition
  block_trs_list :: "'f \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules"
where
  "block_trs_list f trs = map (block_rule f) trs"

lemmas block_defs = block_trs_list_def block_rule_def

definition
  check_superset_of_blocked :: "'f::show \<Rightarrow> ('f, string) rules \<Rightarrow> ('f, string) rules \<Rightarrow> shows check"
where
  "check_superset_of_blocked f R' R \<equiv> (
     check_all (\<lambda>rule. (block_rule f rule) mem R') R
       <+? (\<lambda>rule. ''the rule '' +#+ shows_rule (block_rule f rule)
         +@+ '' is missing'' +#+ shows_nl)
       <+? (\<lambda>e. shows_trs R  +@+ ''is not a subset of'' +#+
         shows_trs R' +@+ e +@+ shows_nl)
  )"

lemma check_superset_of_blocked_sound[simp]:
  assumes "isOK (check_superset_of_blocked f R' R)"
  shows "set (block_trs_list f R) \<subseteq> set R'"
proof (rule subsetI2)
  from assms[unfolded check_superset_of_blocked_def]
    have "forall (\<lambda>r. block_rule f r mem R') R" by simp
  hence a: "\<forall>r\<in>set R. block_rule f r \<in> set R'" unfolding mem_iff by simp
  fix s t assume "(s, t) \<in> set (block_trs_list f R)"
  hence "(s, t) \<in> block_rule f ` set R" unfolding block_defs by simp
  then obtain r where st: "block_rule f r = (s, t)" and "r \<in> set R" by auto
  with a have "block_rule f r \<in> set R'" by blast
  thus "(s, t) \<in> set R'" unfolding st .
qed

definition
  check_left_linear_trs :: "('f::show, string) rules \<Rightarrow> shows check"
where
  "check_left_linear_trs trs \<equiv>
     check (forall (\<lambda>r. linear_term (fst r)) trs) (
       shows trs +@+ shows_nl +@+ ''is not left-linear'' +#+ shows_nl)"

definition
  fcc_proc ::
    "('f::show, string) rules
     \<Rightarrow> ('f, string) rules
     \<Rightarrow> 'f
     \<Rightarrow> ('f, string) ctxt list
     \<Rightarrow> ('f, string) rules
     \<Rightarrow> ('f, string) rules
     \<Rightarrow> shows check"
where
  "fcc_proc P R f fcs P' R' \<equiv> do {
     let ds = defs_list R;
     let fa = (f, Suc 0);
     let Cf = More f [] \<box> [];
     let fcs' = Cf # fcs;
     let vs = vars_trs_list R;
     let fs = list_union (funas_below_trs_list P) (funas_trs_list R);
     let fas = fa # fs;
     check (\<not> fa mem fs) (shows fa +@+ '' is not fresh'' +#+ shows_nl);
     check (\<not> f mem ds) (shows f +@+ '' is defined in the rules'' +#+ shows_nl);
     check_left_linear_trs R;
     check_wf_trs R;
     check_allm (\<lambda>r. check_no_var (fst r) \<guillemotright> check_no_var (snd r) \<guillemotright> check_no_defined_root ds (snd r)) P;
     check_allm (check_flat_ctxt vs) fcs';
     check_allm (check_is_flat_ctxt vs fas) fcs';
     check_allm (check_flat_ctxt_complete fcs') fas;
     check_allm (check_rule_preserving fcs' R') R;
     check_allm (check_rule_reflecting fcs' R) R';
     check_superset_of_blocked f P' P
   }"

fun
  has_unary_root :: "'f \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "has_unary_root f (Fun g [t]) = (f = g)" |
  "has_unary_root f t = False"

fun strip_unary_root :: "'f \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term"
where
  "strip_unary_root f (Fun g [t]) = (if f = g then t else Fun g [t])" |
  "strip_unary_root f t = t"

fun
  unblock_term :: "'f \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term"
where
  "unblock_term f (Fun g ts) = (if forall (has_unary_root f) ts
     then Fun g (map (strip_unary_root f) ts)
     else Fun g ts)" |
  "unblock_term f t = t"

lemma [simp]: "strip_unary_root f \<circ> (\<lambda>t. Fun f [t]) = id"
by (rule ext) (simp add: o_def)

lemma unblock_block_term_ident[simp]: "unblock_term f (block_term f t) = t"
by (cases t) simp_all

definition
  unblock_rule :: "'f \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f, 'v) rule"
where
  "unblock_rule f r \<equiv> (unblock_term f (fst r), unblock_term f (snd r))"

definition
  unblock_trs_list :: "'f \<Rightarrow> ('f, 'v) rules \<Rightarrow> ('f, 'v) rules"
where
  "unblock_trs_list f trs \<equiv> map (unblock_rule f) trs"

lemma unblock_block_trs_list_ident[simp]:
  "unblock_trs_list f (block_trs_list f trs) = trs"
unfolding unblock_trs_list_def block_trs_list_def
by (induct trs) (simp_all add: unblock_rule_def block_rule_def)

lemma [simp]:
  assumes "set (block_trs_list f P) \<subseteq> set P'"
  shows "set P \<subseteq> set (unblock_trs_list f P')"
proof (rule subsetI2)
  fix s t assume "(s, t) \<in> set P"
  hence "block_rule f (s, t) \<in> block_rule f ` set P" by simp
  with assms have "block_rule f (s, t) \<in> set P'" unfolding block_defs by auto
  hence "unblock_rule f (block_rule f (s, t)) \<in> unblock_rule f ` set P'" by simp
  thus "(s, t) \<in> set (unblock_trs_list f P')"
    unfolding block_rule_def unblock_rule_def unblock_trs_list_def by simp
qed

lemma sub_apply_term_block_term_distrib[simp]:
  assumes "\<not> is_var t"
  shows "block_term f t \<cdot> \<sigma> = block_term f (t \<cdot> \<sigma>)"
using assms by (cases t) simp_all

lemma blocked_rstep_imp_rstep:
  assumes refl: "\<forall>(l, r)\<in>R'. \<exists>(u, v)\<in>R. \<exists>C\<in>set (\<box> # fcs). l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
    and rstep: "(s, t) \<in> rstep R'"
  shows "(s, t) \<in> rstep R"
proof -
  from rstep obtain C l r \<sigma> where "(l, r) \<in> R'"
    and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  from refl and `(l, r) \<in> R'`
    have "(l, r) \<in> R \<or> (\<exists>(u, v)\<in>R. \<exists>C\<in>set fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>)" by auto
  thus ?thesis
  proof
    assume "(l, r) \<in> R" thus "(s, t) \<in> rstep R" unfolding s t by auto
  next
    assume "\<exists>(u, v)\<in>R. \<exists>C\<in>set fcs. l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
    then obtain u v D where "(u, v) \<in> R" and "D \<in> set fcs" and l: "l = D\<langle>u\<rangle>" and r: "r = D\<langle>v\<rangle>" by auto
    let ?C = "C \<circ>c (D \<cdot>c \<sigma>)"
    have s': "s = ?C\<langle>u\<cdot>\<sigma>\<rangle>" unfolding s l by simp
    have t': "t = ?C\<langle>v\<cdot>\<sigma>\<rangle>" unfolding t r by simp
    from `(u, v) \<in> R` show "(s, t) \<in> rstep R" unfolding s' t' by best
  qed
qed

lemma check_rules_preserving_sound_spec:
  fixes f::"'f::show" and fcs::"('f, string) ctxt list"
  defines "fcs' \<equiv> More f [] \<box> [] # fcs"
  assumes "isOK (check_allm (check_rule_preserving fcs' R') R)"
  shows "\<forall>(l, r)\<in>set R. (l, r) \<in> set R' \<or> (Fun f [l], Fun f [r]) \<in> set R'"
proof -
  from check_rules_preserving_sound[OF assms(2)]
  show ?thesis unfolding fcs'_def by force
qed

(* --> AbstractRewriting *)
lemmas SN_elt_mono = SN_elt_subset

lemma rstep_preserves_undefined_root:
  assumes "wf_trs R" and "\<not> defined R f" and "(Fun f ss, t) \<in> rstep R"
  shows "\<exists>ts. length ts = length ss \<and> t = Fun f ts"
proof -
  from `wf_trs R` and `(Fun f ss, t) \<in> rstep R` show ?thesis
  proof (cases rule: rstep_cases_Fun')
    case (root ls r \<sigma>)
    with `\<not> defined R f` show ?thesis by (auto simp: defined_def)
  next
    case (nonroot i u) thus ?thesis by simp
  qed
qed

lemma unary_flat_ctxts:
  assumes "(f, Suc 0) \<in> F"
  shows "More f [] \<box> [] \<in> flat_ctxts V F"
proof -
  let ?xs = "map Var (fresh_vars (Suc 0) V)"
  have "flat_ctxt_i ?xs f 0 = More f [] \<box> []"
    unfolding flat_ctxt_i_def fresh_vars_def fresh_strings_def by simp
  with assms show ?thesis unfolding flat_ctxts_def by force
qed

lemma flat_ctxts_funas:
  fixes n::nat and V::"string set"
  defines "xs \<equiv> map Var (fresh_vars n V)"
  assumes "(f, n) \<in> F"
  shows "\<forall>i<n. More f (take i xs) \<box> (drop (Suc i) xs) \<in> flat_ctxts V F"
proof (intro allI impI)
  fix i assume "i < n"
  hence "flat_ctxt_i xs f i = More f (take i xs) \<box> (drop (Suc i) xs)"
    unfolding flat_ctxt_i_def xs_def by simp
  with `i < n`
    have "More f (take i xs) \<box> (drop (Suc i) xs) \<in> {flat_ctxt_i xs f i | i. i \<in> {0..<n}}"
    unfolding xs_def by force
  with assms show "More f (take i xs) \<box> (drop (Suc i) xs) \<in> flat_ctxts V F"
    unfolding flat_ctxts_def xs_def by force
qed

lemma rstep_imp_blocked_rstep:
  fixes R::"('f, string) trs"
  assumes "wf_trs R"
    and flat_ctxts: "\<forall>C\<in>fcs. is_flat_ctxt (vars_trs R) F C"
    and complete: "\<forall>(f, n)\<in>F. \<forall>i<n. \<exists>ss1 ss2.
      (More f ss1 \<box> ss2) \<in> fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
    and "finite R"
    and "funas_below_term s \<subseteq> funas_below_trs P \<union> funas_trs R"
    and "funas_below_trs P \<union> funas_trs R \<subseteq> F"
    and "(f, Suc 0) \<in> F"
    and "\<not> defined R (the(root s))"
    and "(s, t) \<in> rstep R"
    and pres: "\<forall>(l, r)\<in>R. (l, r) \<in> R' \<or> (\<forall>E\<in>fcs. (E\<langle>l\<rangle>, E\<langle>r\<rangle>) \<in> R')"
  shows "\<not> defined R (the(root t)) \<and> funas_below_term t \<subseteq> funas_below_trs P \<union> funas_trs R
    \<and> (block_term f s, block_term f t) \<in> rstep R'" (is "_ \<and> _ \<and> (?s, ?t) \<in> rstep R'")
proof -
  from `(s, t) \<in> rstep R` obtain C l r \<sigma> where "(l, r) \<in>  R"
    and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" by auto
  from `wf_trs R` and `(l, r) \<in> R` obtain g ls
    where l: "l = Fun g ls" by (cases l) (auto simp: wf_trs_def)
  show ?thesis
  proof (cases C)
    case Hole
    have "the(root s) = g" by (simp add: s l Hole)
    with `(l, r) \<in> R` have "defined R (the(root s))" by (auto simp: defined_def l)
    with assms show ?thesis by simp
  next
    case (More h ss1 D ss2)
    note C = this
    let ?block_list = "map (\<lambda>t. Fun f [t])"
    let ?ss1 = "?block_list ss1"
    let ?ss2 = "?block_list ss2"
    let ?D = "More h ?ss1 (More f [] D []) ?ss2"
    have s': "?s = ?D\<langle>l\<cdot>\<sigma>\<rangle>" by (simp add: s More)
    have t': "?t = ?D\<langle>r\<cdot>\<sigma>\<rangle>" by (simp add: t More)
    
    from `finite R` have fin_vars: "finite (vars_trs R)" by (rule finite_trs_has_finite_vars)

    have "the(root s) = h" unfolding s More by simp
    moreover have "the(root t) = h" unfolding t More by simp
    ultimately have no_def: "\<not> defined R (the (root t))" using assms by simp

    have funas_below: "funas_below_term t \<subseteq> funas_below_trs P \<union> funas_trs R"
    proof -
      have "funas_below_term t = \<Union>set (map funas_term (ss1 @ (D\<langle>r \<cdot> \<sigma>\<rangle>) # ss2 ))"
        unfolding t More funas_below_term_def by simp
      moreover have "funas_term (D\<langle>r \<cdot> \<sigma>\<rangle>) \<subseteq> funas_below_trs P \<union> funas_trs R"
      proof -
        from `funas_below_term s \<subseteq> funas_below_trs P \<union> funas_trs R`
          have fs: "funas_term (D\<langle>l \<cdot> \<sigma>\<rangle>) \<subseteq> funas_below_trs P \<union> funas_trs R"
          unfolding s C funas_below_term_def by simp

        have fr: "funas_term (r \<cdot> \<sigma>) = funas_term r \<union> (\<Union>x\<in>vars_term r. funas_term (Var x \<cdot> \<sigma>))"
          by (rule funas_term_sub_apply)
        have fl: "funas_term (l \<cdot> \<sigma>) = funas_term l \<union> (\<Union>x\<in>vars_term l. funas_term (Var x \<cdot> \<sigma>))"
          by (rule funas_term_sub_apply)

        from fs and fl
          have "(\<Union>x\<in>vars_term l. funas_term (Var x \<cdot> \<sigma>)) \<subseteq> funas_below_trs P \<union> funas_trs R"
          unfolding funas_term_ctxt_apply by simp
        moreover from `(l, r) \<in> R` and `wf_trs R`
          have "vars_term r \<subseteq> vars_term l" unfolding wf_trs_def by simp
        ultimately
          have "(\<Union>x\<in>vars_term r. funas_term (Var x \<cdot> \<sigma>)) \<subseteq> funas_below_trs P \<union> funas_trs R"
          by auto
        moreover from `(l, r) \<in> R` have "funas_term r \<subseteq> funas_trs R"
          unfolding funas_defs by auto
        ultimately have "funas_term (r \<cdot> \<sigma>) \<subseteq> funas_below_trs P \<union> funas_trs R"
          unfolding fr by auto
        moreover from fs have "funas_ctxt D \<subseteq> funas_below_trs P \<union> funas_trs R"
          unfolding s More funas_term_ctxt_apply by simp
        ultimately show ?thesis unfolding funas_term_ctxt_apply by simp
      qed
      ultimately show ?thesis using `funas_below_term s \<subseteq> funas_below_trs P \<union> funas_trs R`
        unfolding s C funas_below_term_def by simp
    qed

    from bspec[OF pres `(l, r) \<in> R`, unfolded split_def, simplified] show ?thesis
    proof
      assume "(l, r) \<in> R'" thus ?thesis unfolding s' t' using no_def funas_below by best
    next
      assume flat_ctxt: "\<forall>E\<in>fcs. (E\<langle>l\<rangle>, E\<langle>r\<rangle>) \<in> R'"
      show ?thesis
      proof (cases D rule: ctxt_exhaust_rev)
        case Hole
        let ?E = "More f [] \<box> []"
        let ?F = "More h ?ss1 \<box> ?ss2"

        from `(f, Suc 0) \<in> F` and complete
          have "More f [] \<box> [] \<in> fcs" by auto
        with flat_ctxt
          have "(?E\<langle>l\<rangle>, ?E\<langle>r\<rangle>) \<in> R'" by best
        hence "((?E\<langle>l\<rangle>)\<cdot>\<sigma>, (?E\<langle>r\<rangle>)\<cdot>\<sigma>) \<in> rstep R'" by blast
        hence "(?E\<langle>l\<cdot>\<sigma>\<rangle>, ?E\<langle>r\<cdot>\<sigma>\<rangle>) \<in> rstep R'" by simp
        from rstep.ctxt[OF this]
          have "(?F\<langle>?E\<langle>l\<cdot>\<sigma>\<rangle>\<rangle>, ?F\<langle>?E\<langle>r\<cdot>\<sigma>\<rangle>\<rangle>) \<in> rstep R'" .
        hence "(?s, ?t) \<in> rstep R'" unfolding s' t' Hole by simp
        with no_def and funas_below show ?thesis by simp
      next
        case (More D' f' ss1' ss2')
        let ?n = "Suc (length (ss1' @ ss2'))"
        let ?i = "length ss1'"
        
        from `(l, r) \<in> R`
          have vl: "vars_term l \<subseteq> vars_trs R"
          and vr: "vars_term r \<subseteq> vars_trs R"
          unfolding vars_defs by auto
        
        from coincidence_lemma'[OF vl, symmetric] coincidence_lemma[symmetric]
          have l_sub_R: "l \<cdot> (\<sigma>|_vars_trs R) = l \<cdot> \<sigma>" by simp
        from coincidence_lemma'[OF vr, symmetric] coincidence_lemma[symmetric]
          have r_sub_R: "r \<cdot> (\<sigma>|_vars_trs R) = r \<cdot> \<sigma>" by simp

        from `funas_below_term s \<subseteq> funas_below_trs P \<union> funas_trs R`
          and `funas_below_trs P \<union> funas_trs R \<subseteq> F`
          have "funas_below_term s \<subseteq> F" by blast
        hence in_F: "(f', ?n) \<in> F"
          unfolding funas_below_term_def s C More by simp
        have "?i < ?n" by simp
        with in_F and complete obtain ss1 ss2 where "More f' ss1 \<box> ss2 \<in> fcs" (is "?G \<in> _")
          and "length ss1 = ?i" and "length ss2 = ?n - ?i - 1" by best
        with flat_ctxt have in_R': "(?G\<langle>l\<rangle>, ?G\<langle>r\<rangle>) \<in> R'" by auto

        from flat_ctxts and `?G \<in> fcs` have "is_flat_ctxt (vars_trs R) F ?G" by best
        from is_flat_ctxt_imp_flat_ctxt[OF this]
          have dist_vars: "vars_ctxt ?G \<inter> vars_trs R = {}"
          and "(f', Suc (length (ss1 @ ss2))) \<in> F"
          and all_vars: "forall is_var (ss1 @ ss2)"
          and "distinct (ss1 @ ss2)" by auto

        let ?ss1' = "map the_var ss1"
        let ?ss2' = "map the_var ss2"

        from all_vars have "forall is_var ss1" by simp
        from all_vars have "forall is_var ss2" by simp

        from dist_vars have "\<Union>set(map vars_term ss1) \<inter> vars_trs R = {}" by auto
        with `forall is_var ss1` have "set ?ss1' \<inter> vars_trs R = {}" by (induct ss1) auto
        moreover from dist_vars have "\<Union>set(map vars_term ss2) \<inter> vars_trs R = {}" by auto
        with `forall is_var ss2` have "set ?ss2' \<inter> vars_trs R = {}" by (induct ss2) auto
        ultimately have fresh: "vars_trs R \<inter> set(?ss1' @ ?ss2') = {}" by auto
        
        from Var_the_var_id[OF all_vars]
          have "map Var (map the_var (ss1 @ ss2)) = ss1 @ ss2" .
        from distinct_the_vars[OF all_vars `distinct (ss1 @ ss2)`]
          have "distinct (map the_var (ss1 @ ss2))" .
        hence dist: "distinct (?ss1' @ ?ss2')" by simp
        
        have len1: "length ?ss1' = length ss1'" using `length ss1 = ?i` by simp

        have len2: "length ?ss2' = length ss2'" using `length ss2 = ?n - ?i - 1` by simp
        
        let ?\<sigma> = "sub_extend \<sigma> (zip (?ss1' @ ?ss2') (ss1' @ ss2'))"
        let ?E' = "More f' ss1' \<box> ss2'"
                
        have G_l_subex: "(?G\<langle>l\<rangle>) \<cdot> ?\<sigma> = ?E'\<langle>l \<cdot> \<sigma>\<rangle>"
          unfolding sub_apply_term_ctxt_apply_distrib
          unfolding sub_extend_id[OF fresh vl]
          using sub_extend_flat_ctxt'[OF dist len1 len2]
          unfolding Var_the_var_id[OF `forall is_var ss1`]
          unfolding Var_the_var_id[OF `forall is_var ss2`] by simp
       have G_r_subex: "(?G\<langle>r\<rangle>) \<cdot> ?\<sigma> = ?E'\<langle>r \<cdot> \<sigma>\<rangle>"
          unfolding sub_apply_term_ctxt_apply_distrib
          unfolding sub_extend_id[OF fresh vr]
          using sub_extend_flat_ctxt'[OF dist len1 len2]
          unfolding Var_the_var_id[OF `forall is_var ss1`]
          unfolding Var_the_var_id[OF `forall is_var ss2`] by simp

       from rstep.sub[OF rstep.id[OF in_R'], of ?\<sigma>]
         have "(?E'\<langle>l \<cdot> \<sigma>\<rangle>, ?E'\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep R'"
          unfolding G_l_subex G_r_subex .
        from rstep.ctxt[OF this, of D']
          have "(D\<langle>l \<cdot> \<sigma>\<rangle>, D\<langle>r \<cdot> \<sigma>\<rangle>) \<in> rstep R'"
          unfolding More by simp
        from rstep.ctxt[OF this, of "More h ?ss1 (More f [] \<box> []) ?ss2"]
          have "(?s, ?t) \<in> rstep R'" unfolding s t C by simp
        with no_def and funas_below show ?thesis by simp
      qed
    qed
  qed
qed

lemma rseq_imp_blocked_rseq:
  fixes R::"('f, string) trs"
  assumes "wf_trs R"
    and flat_ctxts: "\<forall>C\<in>fcs. is_flat_ctxt (vars_trs R) F C"
    and complete: "\<forall>(f, n)\<in>F. \<forall>i<n. \<exists>ss1 ss2.
      (More f ss1 \<box> ss2) \<in> fcs \<and> length ss1 = i \<and> length ss2 = n - i - 1"
    and "finite R"
    and "funas_below_term s \<subseteq> funas_below_trs P \<union> funas_trs R"
    and "funas_below_trs P \<union> funas_trs R \<subseteq> F"
    and "(f, Suc 0) \<in> F"
    and "\<not> defined R (the (root s))"
    and "(s, t) \<in> (rstep R)^*"
    and pres: "\<forall>(l, r)\<in>R. (l, r) \<in> R' \<or> (\<forall>E\<in>fcs. (E\<langle>l\<rangle>, E\<langle>r\<rangle>) \<in> R')"
  shows "\<not> defined R (the (root t)) \<and> funas_below_term t \<subseteq> funas_below_trs P \<union> funas_trs R
    \<and> (block_term f s, block_term f t) \<in> (rstep R')^*"
using `(s, t) \<in> (rstep R)^*`
  and `funas_below_term s \<subseteq> funas_below_trs P \<union> funas_trs R`
  and `\<not> defined R (the (root s))`
proof (induct rule: rtrancl_induct)
  case base thus ?case by simp
next
  case (step y z)
  with rstep_imp_blocked_rstep[OF `wf_trs R` flat_ctxts complete `finite R`
    _ `funas_below_trs P \<union> funas_trs R \<subseteq> F` `(f, Suc 0) \<in> F` _ _ pres, of y z]
    have "\<not> defined R (the (root z)) \<and> funas_below_term z \<subseteq> funas_below_trs P \<union> funas_trs R
      \<and> (block_term f y, block_term f z) \<in> rstep R'
      \<and> (block_term f s, block_term f y) \<in> (rstep R')^*" by simp
  thus ?case by auto
qed

lemma funas_below_term_Fun:
  "funas_below_term (Fun f ts \<cdot> \<sigma>) = (\<Union>t\<in>set ts. funas_term (t \<cdot> \<sigma>))"
unfolding funas_below_term_def by simp

lemma funas_below_term_sub_apply:
  assumes "\<not> is_var t"
  shows "funas_below_term (t \<cdot> \<sigma>) = (\<Union>t\<in>set (args t). funas_term (t \<cdot> \<sigma>))"
proof -
  from assms obtain f ts where t: "t = Fun f ts" using not_is_var_imp_Fun by best
  show ?thesis using funas_below_term_Fun unfolding t by simp
qed

lemma block_sub_Fun: "block_term f (Fun g us \<cdot> \<sigma>) = (block_term f (Fun g us)) \<cdot> \<sigma>"
by simp

lemma funas_term_clean_subset: "funas_term (clean_term F t) \<subseteq> F" by (induct t) auto

lemma blocked_rstep_imp_rstep':
  assumes "wf_trs R"
    and "\<not> defined R f"
    and "\<not> defined R (the (root s))"
    and "(block_term f s, block_term f t) \<in> rstep R" (is "(?s, ?t) \<in> _")
  shows "\<not> defined R (the (root t)) \<and> (s, t) \<in> rstep R"
proof -
  from NF_var[OF `wf_trs R`] and assms obtain g ss where s: "s = Fun g ss" by (induct s) auto
  hence s': "?s = Fun g (map (\<lambda>t. Fun f [t]) ss)" by simp
  from assms have "\<not> defined R g" unfolding s by simp
  from `wf_trs R` and `(?s, ?t) \<in> rstep R`[unfolded s'] show ?thesis
  proof (cases rule: rstep_cases_Fun')
    case (root ls r \<sigma>)
    with `\<not> defined R g` show ?thesis unfolding defined_def by auto
  next
    case (nonroot i u)
    let ?C = "More g (take i ss) \<box> (drop (Suc i) ss)"
    from nonroot have "(Fun f [ss!i], u) \<in> rstep R" by simp
    with `wf_trs R` have "\<exists>v. u = Fun f [v] \<and> (ss!i, v) \<in> rstep R"
    proof (cases rule: rstep_cases_Fun')
      case (root ls' r' \<sigma>')
      with assms show ?thesis unfolding defined_def by auto
    next
      case (nonroot j v)
      thus ?thesis by auto
    qed
    then obtain v where u: "u = Fun f [v]" and rstep: "(ss!i, v) \<in> rstep R" by auto
    from rstep.ctxt[OF rstep] have "(?C\<langle>ss!i\<rangle>, ?C\<langle>v\<rangle>) \<in> rstep R" .
    moreover have "s = ?C\<langle>ss!i\<rangle>"
    proof -
      from `i < length (map (\<lambda>t. Fun f [t]) ss)` have "i < length ss" by simp
      from id_take_nth_drop[OF this] have "take i ss @ (ss!i) # drop (Suc i) ss = ss" ..
      thus ?thesis by (simp add: s)
    qed
    moreover have "unblock_term f ?t = ?C\<langle>v\<rangle>"
    proof -
      let ?ss = "take i ss @ v # drop (Suc i) ss"
      let ?ss' = "map (\<lambda>t. Fun f [t]) ?ss"
      have map_simp_aux: "Fun f [v] # map (\<lambda>t. Fun f [t]) (drop (Suc i) ss) = map (\<lambda>t. Fun f [t]) (v # drop (Suc i) ss)" by simp
      have "forall (has_unary_root f) ?ss'" by auto
      hence "unblock_term f (Fun g ?ss') = Fun g ?ss" by simp
      thus ?thesis unfolding nonroot
        unfolding take_map u drop_map
        unfolding map.simps[symmetric]
      unfolding nonroot u
      unfolding take_map drop_map
      unfolding map_simp_aux
      unfolding map_append by simp
    qed
    moreover have "\<not> defined R (the (root t))"
    proof -
     have "the (root (unblock_term f ?t)) = g" unfolding nonroot by simp
     with `\<not> defined R g`
       show ?thesis unfolding unblock_block_term_ident by simp
    qed
    ultimately show ?thesis by simp
  qed
qed

lemma rstep_imp_blocked_rstep':
  assumes "wf_trs R"
    and "\<not> defined R f"
    and "\<not> defined R (the (root s))"
    and "(block_term f s, t) \<in> rstep R"
  shows "\<exists>u. \<not> defined R (the (root u)) \<and> t = block_term f u"
proof -
  let ?s = "block_term f s"
  from assms and NF_var[OF `wf_trs R`] have "\<not> is_var s" by auto
  then obtain g ss where s: "s = Fun g ss" by (induct s) auto
  let ?ss = "map (\<lambda>t. Fun f [t]) ss"
  have s': "?s = Fun g ?ss" by (simp add: s)
  from assms have "\<not> defined R g" unfolding s by simp
  from rstep_preserves_undefined_root[OF `wf_trs R` this `(?s, t) \<in> rstep R`[unfolded s']]
    obtain ts where "length ts = length ?ss" and t: "t = Fun g ts" by auto
  have "\<not> defined R (the (root t))" unfolding t using `\<not> defined R g` by simp
  from `wf_trs R` and `(?s, t) \<in> rstep R`[unfolded s'] show ?thesis
  proof (cases rule: rstep_cases_Fun')
    case (root ls r \<sigma>)
    from assms and root have "\<not> defined R g" unfolding root s by simp
    with `(Fun g ls, r) \<in> R` show ?thesis unfolding defined_def by auto
  next
    case (nonroot i u)
    from assms have "\<not> defined R (the (root t))" unfolding nonroot s by simp
    from nonroot have "(Fun f [ss!i], u) \<in> rstep R" by simp
    from rstep_preserves_undefined_root[OF `wf_trs R` `\<not> defined R f` this]
      obtain ts where "length ts = length [ss!i]" and "u = Fun f ts" by auto
    then obtain v where u: "u = Fun f [v]" by (induct ts) simp_all
    let ?u = "Fun g (take i ss @ v # drop (Suc i) ss)"
    have ss1: "map (\<lambda>t. Fun f [t]) (take i ss) = take i ?ss" unfolding take_map ..
    have ss2: "map (\<lambda>t. Fun f [t]) (drop (Suc i) ss) = drop (Suc i) ?ss" unfolding drop_map ..
    have "map (\<lambda>t. Fun f [t]) (take i ss @ v # drop (Suc i) ss) = take i ?ss @ (Fun f [v]) # drop (Suc i) ?ss" unfolding map_append by (simp add: ss1 ss2)
    hence "t = block_term f ?u" unfolding nonroot u by simp
    moreover have "\<not> defined R (the (root ?u))" using `\<not> defined R (the (root t))`
      unfolding nonroot by simp
    ultimately show ?thesis by best
  qed
qed

lemma SN_elt_imp_blocked_SN_elt:
  assumes "wf_trs R"
    and "\<not> defined R f"
    and "\<not> SN_elt (rstep R') (block_term f t)"
    and "\<not> defined R (the (root t))"
    and refl: "\<forall>(l, r)\<in>R'. \<exists>(u, v)\<in>R. \<exists>C\<in>set (\<box> # fcs). l = C\<langle>u\<rangle> \<and> r = C\<langle>v\<rangle>"
  shows "\<not> SN_elt (rstep R) t"
proof -
  from assms(3) obtain S where "S 0 = block_term f t"
    and seq: "\<forall>i. (S i, S (Suc i)) \<in> rstep R'" by auto
  from seq and blocked_rstep_imp_rstep[OF refl]
    have R_seq: "\<forall>i. (S i, S (Suc i)) \<in> rstep R" by auto
  have "\<forall>i. \<exists>u. \<not> defined R (the (root u)) \<and> S i = block_term f u"
  proof
    fix i show "\<exists>u. \<not> defined R (the (root u)) \<and> S i = block_term f u"
    proof (induct i)
      case 0
      from `S 0 = block_term f t` and assms show ?case by auto
    next
      case (Suc i)
      then obtain u where "\<not> defined R (the (root u))" and Si: "S i = block_term f u" by auto
      from R_seq have "(S i, S (Suc i)) \<in> rstep R" by simp
      hence "(block_term f u, S (Suc i)) \<in> rstep R" unfolding Si by simp
      from rstep_imp_blocked_rstep'[OF `wf_trs R` `\<not> defined R f`
        `\<not> defined R (the (root u))` this]
        show ?case .
    qed
  qed
  from choice[OF this] obtain T
    where "\<forall>i. \<not> defined R (the (root (T i))) \<and> S i = block_term f (T i)" 
    (is "\<forall>i. _ \<and> S i = ?T i") by auto
  with R_seq have blocked_seq: "\<forall>i. (?T i, ?T (Suc i)) \<in> rstep R" by auto
  let ?T' = "\<lambda>i. unblock_term f (S i)"
  have "?T' 0 = t" unfolding `S 0 = block_term f t` by simp
  moreover have "\<forall>i. (?T' i, ?T' (Suc i)) \<in> rstep R"
  proof
    fix i show "(?T' i, ?T' (Suc i)) \<in> rstep R"
    proof -
      from `\<forall>i. \<not> defined R (the (root (T i))) \<and> S i = ?T i`
        have "\<not> defined R (the (root (T i)))"
        and si: "S i = block_term f (T i)"
        and ssi: "S (Suc i) = block_term f (T (Suc i))" by auto
      from blocked_seq have "(?T i, ?T (Suc i)) \<in> rstep R" ..
      from blocked_rstep_imp_rstep'[OF `wf_trs R` `\<not> defined R f` `\<not> defined R (the (root (T i)))` this]
      show ?thesis unfolding si ssi by simp
    qed
  qed
  ultimately show ?thesis unfolding SN_defs by auto
qed

lemma fcc_proc_sound:
  "sound_proc_impl (\<lambda>P R rm P' R'. isOK (fcc_proc P R f fcs P' R'))"
unfolding sound_proc_impl_def
proof (intro allI impI)
  fix P R P' R'
  assume ok: "isOK (fcc_proc P R f fcs P' R')"
    and finite: "finite_dpp (set P', set R')"
  let ?F' = "list_union (funas_below_trs_list P) (funas_trs_list R)"
  let ?F = "(f, Suc 0) # funas_below_trs_list P @ funas_trs_list R"
  let ?c = "clean_term (set (funas_below_trs_list P @ funas_trs_list R))"
  let ?cs = "clean_sub (set (funas_below_trs_list P @ funas_trs_list R))"
  let ?Cf = "More f [] \<box> []"

   from ok[unfolded fcc_proc_def Let_def]
     have "isOK (check_allm (check_is_flat_ctxt (vars_trs_list R) ((f, Suc 0) # ?F')) (?Cf # fcs))" by auto
     hence "\<forall>C\<in>set (?Cf # fcs). is_flat_ctxt_list (vars_trs_list R) ((f, Suc 0) # ?F') C"
       using check_is_flat_ctxt_def[of "vars_trs_list R" "((f, Suc 0) # ?F')"] by auto
     hence flat_ctxts: "\<forall>C\<in>set (?Cf # fcs). is_flat_ctxt (vars_trs (set R)) (set ?F) C" unfolding is_flat_ctxt_list_sound by auto

  from ok[unfolded fcc_proc_def Let_def]
    check_flat_ctxt_complete_sound[of "?Cf # fcs" ?F]
    have complete: "\<forall>(g, n)\<in>set ?F. \<forall>i<n. \<exists>ss1 ss2.
      More g ss1 \<box> ss2 \<in> set (?Cf # fcs) \<and> length ss1 = i \<and> length ss2 = n - i - 1"
    by auto

  from ok[unfolded fcc_proc_def Let_def] have "wf_trs (set R)" by simp

  from ok[unfolded fcc_proc_def Let_def] have "\<not> f mem defs_list R" by simp
  hence "\<not> defined (set R) f" unfolding defs_list_sound .

  from ok[unfolded fcc_proc_def Let_def]
    check_rules_preserving_sound[of "?Cf # fcs" R' R]
    have pres: "\<forall>(l, r)\<in>set R. (l, r) \<in> set R'
      \<or> (\<forall>C\<in>set (?Cf # fcs). (C\<langle>l\<rangle>, C\<langle>r\<rangle>) \<in> set R')" by auto

  from ok[unfolded fcc_proc_def Let_def]
    check_rules_reflecting_sound[of "?Cf # fcs" R R']
    have refl: "\<forall>(l', r')\<in>set R'. \<exists>(l, r)\<in>set R. \<exists>C\<in>set (\<box> # ?Cf # fcs). l' = C\<langle>l\<rangle> \<and> r' = C\<langle>r\<rangle>" by simp

  from ok[unfolded fcc_proc_def Let_def] check_rules_preserving_sound_spec[of f fcs R' R]
    have "\<forall>(l, r)\<in>set R. (l, r) \<in> set R' \<or> (Fun f [l], Fun f [r]) \<in> set R'" by simp
  show "finite_dpp (set P, set R)"
  proof (rule ccontr)
    assume "\<not> finite_dpp (set P, set R)"
    then obtain s t \<sigma>
      where mchain: "min_ichain (set P, set R) s t \<sigma>" by (auto simp: finite_dpp_def)
    
    from mchain have P_steps: "\<forall>i. (s i, t i) \<in> set P" by simp
    
    have "\<forall>i. \<not> is_var (s i) \<and> \<not> is_var (t i)"
    proof
      fix i
      from P_steps have "(s i, t i) \<in> set P" ..
      with ok[unfolded fcc_proc_def Let_def]
        show "\<not> is_var (s i) \<and> \<not> is_var (t i)" by auto
    qed
    hence si_Fun: "\<forall>i. \<not> is_var (s i)" and ti_Fun: "\<forall>i. \<not> is_var (t i)" by auto

    have funas_below: "funas_below_trs (set P) \<union> funas_trs (set R) \<subseteq> set ?F" by auto

    from ok[unfolded fcc_proc_def Let_def]
      have "left_linear_trs (set R)"
      by (auto simp: left_linear_trs_def check_left_linear_trs_def)
    moreover
      have "funas_below_trs (set P) \<union> funas_trs (set R)
        \<subseteq> set (funas_below_trs_list P @ funas_trs_list R)" by simp
    moreover {
      fix s t assume "(s, t) \<in> set P"
      with ok[unfolded fcc_proc_def Let_def]
        have "\<not> is_var s" and "\<not> is_var t" by auto
      moreover have "\<not> defined (set R) (the (root t))"
      proof -
        from `(s, t) \<in> set P` and ok[unfolded fcc_proc_def Let_def]
          have "\<not> the (root t) mem defs_list R" unfolding defs_list_sound[symmetric] by auto
        thus ?thesis unfolding defs_list_sound[symmetric] .
      qed
      ultimately have "\<not> is_var s \<and> \<not> is_var t \<and> \<not> defined (set R) (the (root t))" by simp
    }
    ultimately
      have clean_mchain: "min_ichain (set P, set R) s t (\<lambda>i. ?cs (\<sigma> i))"
      using mchain by (rule left_linear_min_ichain_imp_clean_min_ichain_fcc)

    let ?s = "\<lambda>i. block_term f (s i)"
    let ?t = "\<lambda>i. block_term f (t i)"

    have "\<forall>i. (?s i, ?t i) \<in> set P'"
    proof
      fix i
      from mchain have "(s i, t i) \<in> set P" by simp
      hence "block_rule f (s i, t i) \<in> block_rule f ` set P" by simp
      hence "(?s i, ?t i) \<in> block_rule f ` set P" unfolding block_defs by simp
      hence "(?s i, ?t i) \<in> set (block_trs_list f P)" unfolding block_defs by simp
      moreover from ok[unfolded fcc_proc_def Let_def]
        have "set (block_trs_list f P) \<subseteq> set P'"
        using check_superset_of_blocked_sound by auto
      ultimately show "(?s i, ?t i) \<in> set P'" by blast
    qed

    moreover have "\<forall>i. (?t i \<cdot> ?cs (\<sigma> i), ?s (Suc i) \<cdot> ?cs (\<sigma> (Suc i))) \<in> (rstep (set R'))^*"
      (is "\<forall>i. ?P i")
    proof
      fix i
      let ?\<sigma> = "?cs (\<sigma> i)"
      let ?t = "t i \<cdot> ?cs (\<sigma> i)"
      let ?\<sigma>' = "?cs (\<sigma> (Suc i))"
      let ?s = "s (Suc i) \<cdot> ?\<sigma>'"

      from clean_mchain have t_in_P: "(s i, t i) \<in> set P" by simp
      from P_steps have s_in_P: "(s (Suc i), t (Suc i)) \<in> set P" by simp
      from clean_mchain have R_seq: "(?t, ?s) \<in> (rstep (set R))^*" by simp

      from ti_Fun have "\<not> is_var (t i)" by simp
      then obtain gs ts where ti: "t i = Fun gs ts" by best
      
      from si_Fun have "\<not> is_var (s (Suc i))" by simp
      then obtain hs us where si: "s (Suc i) = Fun hs us" by best

      from t_in_P and ok[unfolded fcc_proc_def Let_def]
        have "\<not> the (root (t i)) mem defs_list R" unfolding defs_list_sound[symmetric] by auto
      hence "\<not> defined (set R) (the (root (t i)))"
        unfolding defs_list_sound .
      hence not_def_t: "\<not> defined (set R) (the (root ?t))" unfolding ti by simp
  
      have t_funas_below: "funas_below_term ?t \<subseteq> funas_below_trs (set P) \<union> funas_trs (set R)"
      proof -
        from t_in_P
          have funas_ti: "funas_below_term (t i) \<subseteq> funas_below_trs (set P) \<union> funas_trs (set R)"
          unfolding funas_below_trs_def funas_below_rule_def by auto
        moreover have "funas_below_term ?t = (\<Union>t\<in>set ts. funas_term (t \<cdot> ?\<sigma>))"
          unfolding funas_below_term_sub_apply[OF `\<not> is_var (t i)`]
          unfolding ti by auto
        moreover
          have "\<forall>t\<in>set ts. funas_term (t \<cdot> ?\<sigma>) \<subseteq> funas_below_trs (set P) \<union> funas_trs (set R)"
        proof
          fix u assume "u \<in> set ts"
          from funas_ti[unfolded ti funas_below_term_def]
            have "\<Union>set (map funas_term ts)
              \<subseteq> funas_below_trs (set P) \<union> funas_trs (set R)" by simp
          with `u \<in> set ts` have "funas_term u
            \<subseteq> funas_below_trs (set P) \<union> funas_trs (set R)" by auto
          thus "funas_term (u \<cdot> ?\<sigma>)
            \<subseteq> funas_below_trs (set P) \<union> funas_trs (set R)"
          proof (induct u)
            case (Var x)
            have "funas_term (?c (get_sub (\<sigma> i) x))
              \<subseteq> set (funas_below_trs_list P @ funas_trs_list R)"
              by (simp add: funas_term_clean_subset)
            hence "funas_term (Var x \<cdot> ?cs (\<sigma> i))
              \<subseteq> set (funas_below_trs_list P @ funas_trs_list R)" by (cases "\<sigma> i") simp_all
            thus ?case by simp
          next
            case (Fun i us) thus ?case by auto
          qed
        qed
        ultimately show ?thesis by auto
      qed
      from rseq_imp_blocked_rseq[OF `wf_trs (set R)` flat_ctxts complete finite_set[of R]
        this funas_below _ not_def_t R_seq pres]
      have "(block_term f ?t, block_term f ?s) \<in> (rstep (set R'))^*" by simp
      thus "?P i" unfolding si ti unfolding block_sub_Fun .
    qed

    moreover have "\<forall>i. SN_elt (rstep (set R')) (?t i \<cdot> ?cs (\<sigma> i))"
    proof
      fix i
      from ti_Fun obtain k vs where ti: "t i = Fun k vs" by best
      have "the (root (t i)) = k" unfolding ti by simp
      hence k: "k = the (root (t i \<cdot> ?cs (\<sigma> i)))" unfolding ti by simp
      from P_steps have "(s i, t i) \<in> set P" by simp
      from ok[unfolded fcc_proc_def Let_def]
        have "\<forall>(s, t)\<in>set P. \<not> the (root t) mem (defs_list R)"
        unfolding defs_list_sound[symmetric] by auto
      with `(s i, t i) \<in> set P` have "\<not> defined (set R) (the (root (t i)))"
        unfolding defs_list_sound by auto
      hence "\<not> defined (set R) k" unfolding `the (root (t i)) = k` .
      hence not_def_t: "\<not> defined (set R) (the (root (t i \<cdot> ?cs (\<sigma> i))))" unfolding k .
      from block_sub_Fun[of f k vs "?cs (\<sigma> i)"] have block_ti: "?t i \<cdot> ?cs (\<sigma> i) = block_term f (t i \<cdot> ?cs (\<sigma> i))" unfolding ti by auto
      from clean_mchain have snelt: "SN_elt (rstep (set R)) (t i \<cdot> ?cs (\<sigma> i))" unfolding min_ichain.simps minimal_cond_def by best
      show "SN_elt (rstep (set R')) (?t i \<cdot> ?cs (\<sigma> i))"
      proof (rule ccontr)
        assume "\<not> SN_elt (rstep (set R')) (?t i \<cdot> ?cs (\<sigma> i))"
        hence "\<not> SN_elt (rstep (set R')) (block_term f (t i \<cdot> ?cs (\<sigma> i)))" unfolding block_ti by simp
        from SN_elt_imp_blocked_SN_elt[OF `wf_trs (set R)` `\<not> defined (set R) f` this not_def_t refl] snelt
        show False by simp
      qed
    qed

    ultimately have "min_ichain (set P', set R') ?s ?t (\<lambda>i. ?cs (\<sigma> i))"
      by (auto simp: minimal_cond_def)
    with finite show False unfolding finite_dpp_def by best
  qed
qed


(* Root-Labeling *)

locale rl_interpr =
  fixes F::"('f::show,'l::show)lab set"
    and f::"('f,'l)lab"
  assumes ne: "f \<in> F"
begin
fun rl_inter
where "rl_inter g _ = (if g \<in> F then g else f)"
end

locale rl_interpr_impl =
  fixes fs::"('f::show,'l::show)lab list"
    and f::"('f,'l)lab"
  assumes ne: "f mem fs"

sublocale rl_interpr_impl \<subseteq> rl_interpr "set fs"
proof
  from ne show "f \<in> set fs" unfolding mem_iff .
qed

fun rl_inter_list
where "rl_inter_list fs d g _  = (if g mem fs then g else d)"

context rl_interpr_impl
begin
lemma ne2: "f \<in> set fs" using ne unfolding mem_iff by simp

lemma rl_inter_list_sound[simp]:
  shows  "rl_inter_list fs f = rl_inter"
proof (intro ext)
  fix g cs show "rl_inter_list fs f g cs = rl_inter g cs" by (cases "g mem fs") (auto simp: mem_iff)
qed
end

sublocale rl_interpr \<subseteq> sl_interpr "F" "f" "rl_inter" "{}" L
proof (unfold_locales,rule ne)
  show "SN(rstep(decr_of_ord{}))" unfolding decr_of_ord_def by simp
  show "wf_inter rl_inter F" unfolding wf_inter_def
  proof (intro allI impI)
    fix f cs assume "set cs \<subseteq> F"
    show "rl_inter f cs \<in> F" using ne by auto
  qed
qed

context rl_interpr_impl
begin

lemma sl_interpr_impl: "sl_interpr_impl  (set fs) f (rl_inter_list fs f) {} L
  (check_sl_lab_trs (rl_inter_list fs f) L fs)
  (check_sl_model_lab_trs (rl_inter_list fs f) L fs)"
by (unfold_locales,rule ne2,simp,rule wf_I,rule Decr_SN,
  simp only: check_sl_lab_trs_sound,
  simp only: check_sl_model_lab_trs_sound,auto)
end

definition rl_fun where "rl_fun _ cs = Inr cs"

definition
  rl_tt :: "(('f::show, 'l::show) lab, string) rules  \<Rightarrow> (('f, 'l) lab, string) rules \<Rightarrow> shows check"
where
  "rl_tt R lR \<equiv> do {
     let fs = funs_trs_list R;
     check (fs \<noteq> [])
       (''root-labeling requires at least one function symbol in the signature'' +#+ shows_nl);
     let d = hd fs;
     check_sem_lab_tt
       (check_sl_lab_trs (rl_inter_list fs d) rl_fun fs)
       (check_sl_model_lab_trs (rl_inter_list fs d) rl_fun fs)
       {}
       (rl_inter_list fs d) rl_fun R lR
   }"

(* The Root-Labeling Processor *)
definition
  rl_proc ::
    "(('f::show,'l::show)lab,string)rules
     \<Rightarrow> (('f,'l)lab,string)rules \<Rightarrow> (('f,'l)lab,string)rules \<Rightarrow> (('f,'l)lab,string)rules
     \<Rightarrow> shows check"
where
  "rl_proc P R lP lR \<equiv> do {
     let fs = list_union (funs_trs_list R) (list_diff (funs_trs_list P) (defs_list P));
     check (fs \<noteq> [])
       (''root-labeling requires at least one function symbol in the signature'' +#+ shows_nl);
     let d = hd fs;
     check_sem_lab_proc
       (check_sl_lab_trs (rl_inter_list fs d) rl_fun fs)
       (check_sl_model_lab_trs (rl_inter_list fs d) rl_fun fs)
       {}
       (rl_inter_list fs d) rl_fun P R lP lR
   }"


lemma rl_tt_sound: "sound_lab_tt_impl(\<lambda>R m lR rm. rm = Suc m \<and> isOK(rl_tt R lR))"
unfolding sound_lab_tt_impl_def
proof (intro allI impI)
  fix R::"(('f::show,'l::show)lab,string)rules" and m rR rm
  assume check: "rm = Suc m \<and> isOK(rl_tt R rR)" and decr: "decrSN (set rR) rm"
  let ?fs = "funs_trs_list R"
  let ?def = "hd ?fs"
  from check have "?fs \<noteq> []" unfolding rl_tt_def by auto
  hence "?def mem ?fs" by (cases "?fs") simp_all
  hence rl: "rl_interpr_impl ?fs ?def" unfolding rl_interpr_impl_def rl_interpr_def .
  have rm: "rm = Suc m" using check by simp
  show "decrSN (set R) m"
    using sl_interpr_impl.check_sem_lab_tt_sound[OF rl_interpr_impl.sl_interpr_impl[OF rl,of rl_fun]] check
    unfolding sound_lab_tt_impl_def rl_tt_def
    using rm decr by (auto simp: Let_def)
qed

lemma rl_proc_sound: "sound_lab_proc_impl(\<lambda>P R m ruleMap lP lR rm. rm = Suc m \<and> isOK(rl_proc P R lP lR))"
unfolding sound_lab_proc_impl_def
proof (intro allI impI)
  fix P R::"(('f::show,'l::show)lab,string)rules" and m lP lR rm
  assume check: "rm = Suc m \<and> isOK(rl_proc P R lP lR)" and finite: "finite_lab_dpp(set lP,set lR) rm"
  let ?fs = "list_union (funs_trs_list R) (list_diff (funs_trs_list P) (defs_list P))"
  let ?def = "hd ?fs"
  from check have "?fs \<noteq> []" unfolding rl_proc_def by auto
  hence ne: "?def mem ?fs" by (cases "?fs") simp_all
  hence rl: "rl_interpr_impl ?fs ?def" unfolding rl_interpr_impl_def rl_interpr_impl_def .
  have rm: "rm = Suc m" using check by simp
  show "finite_lab_dpp(set P,set R) m"
    using sl_interpr_impl.check_sem_lab_proc_sound[OF rl_interpr_impl.sl_interpr_impl[OF rl,of rl_fun]] check
    unfolding sound_lab_proc_impl_def rl_proc_def Let_def
    using rm finite by force
qed

end
