(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory StringReversal imports DpFramework begin

fun
  unary_term :: "('f, 'v) term \<Rightarrow> bool"
where
  "unary_term(Var x) = True" |
  "unary_term(Fun f [t]) = unary_term t" |
  "unary_term t = False"

lemma unary_term_single_arg: assumes "unary_term(Fun f ts)" shows "\<exists>s. ts = [s]"
proof -
  from assms obtain t where t: "t = Fun f ts" by simp
  moreover with assms have "unary_term t" by simp
  ultimately show ?thesis by (induct rule: unary_term.induct) auto
qed

lemma unary_term_induct[consumes 1,case_names Var Fun,induct type: "term"]:
  fixes P :: "('f,'v)term \<Rightarrow> bool"
  assumes "unary_term t"
    and "\<And>x. P(Var x)"
    and "\<And>f s. unary_term s \<Longrightarrow> P s \<Longrightarrow> P(Fun f [s])"
  shows "P t"
using assms(1) proof (induct t)
  case (Var x) show ?case by (simp add: assms)
next
  case (Fun f ss)
  with unary_term_single_arg obtain s where [simp]: "ss = [s]" by best
  from Fun have "unary_term s" by simp
  moreover with Fun have "P s" by simp
  ultimately show ?case using assms by simp
qed

fun unary_ctxt :: "('f,'v)ctxt \<Rightarrow> bool"
where "unary_ctxt \<box> = True"
    | "unary_ctxt(More _ [] C []) = unary_ctxt C"
    | "unary_ctxt _ = False"

lemma unary_ctxt_single_arg: assumes "unary_ctxt(More f ss1 D ss2)" shows "ss1 = [] \<and> ss2 = []"
proof -
  from assms obtain E where "E = More f ss1 D ss2" by simp
  moreover with assms have "unary_ctxt E" by simp
  ultimately show ?thesis by (induct rule: unary_ctxt.induct) auto
qed

lemma unary_ctxt_induct[consumes 1,case_names Hole More,induct type: "ctxt"]:
  fixes P :: "('f,'v)ctxt \<Rightarrow> bool"
  assumes "unary_ctxt C"
    and "P \<box>"
    and "\<And>f D. unary_ctxt D \<Longrightarrow> P D \<Longrightarrow> P(More f [] D [])"
  shows "P C"
using assms(1) proof (induct C)
  case Hole show ?case by (simp add: assms)
next
  case (More f ss1 D ss2)
  from unary_ctxt_single_arg[OF More(2)] have [simp]: "ss1 = []" and [simp]: "ss2 = []" by auto
  from More have "unary_ctxt D" by simp
  moreover with More have "P D" by simp
  ultimately show ?case using assms by simp
qed

fun rev_term :: "('f,'v)ctxt \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
where "rev_term C (Var x) = C\<langle>Var x\<rangle>"
    | "rev_term C (Fun f [t]) = rev_term (More f [] C []) t"

fun rev_ctxt :: "('f,'v)ctxt \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)ctxt"
where "rev_ctxt C \<box> = C"
    | "rev_ctxt C (More f [] D []) = rev_ctxt (More f [] C []) D"
    | "rev_ctxt C D = D \<circ>c C" (* the last case is not relevant but it is helpful to be defined *)

fun get_var :: "('f,'v)term \<Rightarrow> 'v"
where "get_var(Var x) = x"
    | "get_var(Fun _ [t]) = get_var t"

definition sub :: "'v \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)sub"
where "sub x C = Sub(\<lambda>y. if x = y then C\<langle>Var x\<rangle> else Var y)"

lemma rev_term_sub[simp]:
  assumes "unary_term t"
    and "unary_ctxt C"
  shows "rev_term C t \<cdot> sub (get_var t) D = rev_term (C \<circ>c D) t"
using assms proof (induct t arbitrary: C rule: unary_term_induct)
  case (Var x) thus ?case by (induct C) (simp_all add: sub_def)
next
  case (Fun f s) thus ?case by simp
qed

lemma rev_term_ctxt_apply[simp]:
  fixes t :: "('f,'v)term"
  assumes "unary_ctxt D"
    and "unary_term t"
  shows "rev_term C (D\<langle>t\<rangle>) = rev_term \<box> t \<cdot> sub (get_var t) (rev_ctxt C D)"
using assms proof (induct D arbitrary: t C rule: unary_ctxt_induct)
  case Hole thus ?case
  proof (induct t rule: unary_term_induct)
    case (Var x) thus ?case by (simp add: sub_def)
  next
    case (Fun f s) thus ?case by simp
  qed
next
  case (More f D)
  have "rev_term C ((More f [] D [])\<langle>t\<rangle>) = rev_term C (Fun f [D\<langle>t\<rangle>])" by simp
  also have "... = rev_term (More f [] C []) (D\<langle>t\<rangle>)" by simp
  also have "... = rev_term \<box> t \<cdot> sub (get_var t) (rev_ctxt (More f [] C []) D)"
    by (simp add: More)
  also have "... = rev_term \<box> t \<cdot> sub (get_var t) (rev_ctxt C (More f [] D []))" by simp
  finally show ?case by simp
qed

lemma rev_ctxt_assoc:
  assumes "unary_ctxt D"
    and "unary_ctxt C"
  shows "rev_ctxt C D = rev_ctxt \<box> D \<circ>c C"
using assms proof (induct D arbitrary: C rule: unary_ctxt_induct)
  case Hole thus ?case 
  proof (induct C rule: unary_ctxt_induct)
    case Hole show ?case by simp
  next
    case (More f C') thus ?case by simp
  qed
next
  case (More f D')
  from More(3) have "unary_ctxt (More f [] C [])" (is "unary_ctxt ?C") by simp
  from More(2)[OF this] have IH: "rev_ctxt ?C D' = rev_ctxt \<box> D' \<circ>c ?C" by simp
  have "unary_ctxt (More f [] \<box> [])" by simp
  from More(2)[OF this]
    have IH': "rev_ctxt (More f [] \<box> []) D' = rev_ctxt \<box> D' \<circ>c More f [] \<box> []" by simp
  have "rev_ctxt C (More f [] D' []) = rev_ctxt ?C D'" by simp
  also have "... = rev_ctxt \<box> D' \<circ>c ?C" unfolding IH by simp
  also have "... = rev_ctxt \<box> D' \<circ>c More f [] \<box> [] \<circ>c C" by simp
  also have "... = rev_ctxt (More f [] \<box> []) D' \<circ>c C" unfolding IH'[symmetric] by simp
  also have "... = rev_ctxt \<box> (More f [] D' []) \<circ>c C" by simp
  finally show ?case by simp
qed

lemma sub_rev_ctxt[simp]:
  assumes "unary_ctxt D" and "unary_ctxt C"
  shows "Var x \<cdot> sub x (rev_ctxt C D) = (rev_ctxt \<box> D)\<langle>C\<langle>Var x\<rangle>\<rangle>"
proof -
  have "rev_ctxt C D = rev_ctxt \<box> D \<circ>c C" by (rule rev_ctxt_assoc[OF assms])
  hence "(rev_ctxt C D)\<langle>Var x\<rangle> = (rev_ctxt \<box> D \<circ>c C)\<langle>Var x\<rangle>" by simp
  thus ?thesis by (auto simp: sub_def)
qed

fun term_to_ctxt :: "('f,'v)term \<Rightarrow> ('f,'v)ctxt"
where "term_to_ctxt(Var _) = \<box>"
    | "term_to_ctxt(Fun f [t]) = More f [] (term_to_ctxt t) []"

fun ctxt :: "'v \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,'v)ctxt"
where "ctxt x (Sub s) = term_to_ctxt (s x)"

lemma unary_ctxt_usb_id[simp]:
  assumes "unary_ctxt C" shows "C \<cdot>c \<sigma> = C"
using assms by (induct C rule: unary_ctxt_induct) simp_all

fun rev_rule :: "('f,'v)rule \<Rightarrow> ('f,'v)rule"
where "rev_rule(l,r) = (rev_term \<box> l,rev_term \<box> r)"

definition rev_trs :: "('f,'v)trs \<Rightarrow> ('f,'v)trs"
where "rev_trs R \<equiv> {rev_rule r | r. r \<in> R}"

definition unary_sig :: "('f\<times>nat)set \<Rightarrow> bool"
where "unary_sig F \<equiv> \<forall>(f,i)\<in>F. i = 1"

lemma unary_ctxt_appl_imp_unary_ctxt : 
  assumes "unary_term C\<langle>t\<rangle>"
  shows "unary_ctxt C"
using assms
proof (induct C, simp)
  case (More f l C r)
  thus ?case
    by (cases l, cases r, simp, simp, cases r, simp, cases "tl l", simp, simp, cases "tl l", simp, simp)
qed

lemma unary_ctxt_appl_imp_unary_term : 
  assumes "unary_term C\<langle>t\<rangle>"
  shows "unary_term t"
using assms
proof (induct C, simp)
  case (More f l C r)
  thus ?case
    by (cases l, cases r, simp, simp, cases r, simp, cases "tl l", simp, simp, cases "tl l", simp, simp)
qed

lemma unary_sub_appl_imp_unary_term :
  assumes "unary_term (t \<cdot> \<sigma>)"
  shows "unary_term t"
using assms 
proof (induct t, simp)
  case (Fun f ss)
  from Fun(2) obtain s where ss: "ss = [s]" by (cases ss, simp, cases "tl ss", simp, simp)
  with Fun(2) have "unary_term (s \<cdot> \<sigma>)" by auto
  with Fun ss show ?case by simp
qed

lemma unary_sub_appl_imp_unary_sub :
  assumes "unary_term (t \<cdot> \<sigma>)"
  shows "unary_term (get_sub \<sigma> (get_var t))"
using assms 
proof (induct t)
  case (Var t)
  thus ?case by (cases \<sigma>, auto)
next
  case (Fun f ss)
  from Fun(2) obtain s where ss: "ss = [s]" by (cases ss, simp, cases "tl ss", simp, simp)
  with Fun(2) have "unary_term (s \<cdot> \<sigma>)" by auto
  with Fun ss show ?case by simp
qed

lemma unary_term_imp_unary_ctxt :
  assumes "unary_term t"
  shows "unary_ctxt (term_to_ctxt t)"
using assms
by (induct rule: unary_term_induct, auto)

lemma unary_rev_ctxt :
  "unary_ctxt (rev_ctxt C D) = (unary_ctxt C \<and> unary_ctxt D)"
proof (induct D arbitrary: C, simp)
  case (More f l D r C)
  thus ?case 
    by (cases l, cases r, simp+)
qed

lemma get_var_sub: 
  assumes string: "unary_term t"  
  shows "get_var (t \<cdot> \<sigma>) = get_var (get_sub \<sigma> (get_var t))"
proof (cases \<sigma>)
  case (Sub sigma)
  from assms
  have "get_var (t \<cdot> Sub sigma) = get_var (sigma (get_var t))"
    by (induct rule: unary_term_induct, auto)
  with Sub show ?thesis by simp
qed
  
lemma get_var_ctxt : assumes "unary_ctxt C"  
  shows "get_var (C\<langle>t\<rangle>) = get_var t"
using assms by (induct rule: unary_ctxt_induct, auto)

lemma get_var_rev_term : assumes "unary_term t"
  and "unary_ctxt C"
  shows "get_var (rev_term C t) = get_var t"
using assms
proof (induct arbitrary: C rule: unary_term_induct)
  case (Var x C)
  let ?x = "Var x"
  have "get_var (rev_term C (?x)) = get_var (C\<langle>?x\<rangle>)" by simp
  also have "\<dots> = get_var (?x)" by (simp add: get_var_ctxt[OF Var])
  finally show ?case .
next
  case (Fun f s C)
  thus ?case by auto
qed

lemma rev_term_sub_apply:
  assumes ut: "unary_term t"
  and uC: "unary_ctxt C"
  shows "rev_term C (t \<cdot> \<sigma>) = rev_term (rev_ctxt C (term_to_ctxt t)) (get_sub \<sigma> (get_var t))"
using assms
proof (induct arbitrary: C rule: unary_term_induct)
  case (Fun f s C)
  thus ?case by simp
next
  case (Var x C)
  thus ?case by (cases \<sigma>, simp)
qed

lemma rev_term_to_rev_ctxt: 
  assumes ut: "unary_term t"
  and uC: "unary_ctxt C"
  shows "rev_term C t = ((rev_ctxt \<box> (term_to_ctxt t)) \<circ>c C)\<langle>Var (get_var t)\<rangle>"
using assms
proof (induct arbitrary: C rule: unary_term_induct)
  case (Var x C)
  thus ?case by auto
  case (Fun f s C)
  from Fun(1) have us: "unary_ctxt (term_to_ctxt s)" by (induct rule: unary_term_induct, auto)
  have uf: "unary_ctxt (More f [] \<box> [])" by simp
  from Fun show ?case by (simp add: rev_ctxt_assoc[OF us uf])
qed
    
lemma rev_ctxt_to_rev_term : 
  assumes "unary_term t"
  and "unary_ctxt C"
  shows "(rev_ctxt C (term_to_ctxt t))\<langle>Var (get_var (get_sub \<sigma> (get_var t)))\<rangle> =
    rev_term C t \<cdot> Sub (\<lambda> x. Var (get_var (get_sub \<sigma> (get_var t))))"
using assms
proof (induct arbitrary: C rule: unary_term_induct)
  case (Fun f s C) thus ?case by auto
next
  case (Var x C) thus ?case by (cases \<sigma>, simp)
qed

lemma rstep_imp_rev_rstep:
  assumes string_s: "unary_term s"
    and string_t: "unary_term t"
    and same_vars: "\<forall> (l,r) \<in> R. get_var l = get_var r"
    and step: "(s,t) \<in> rstep R"
  shows "(rev_term \<box> s,rev_term \<box> t) \<in> rstep(rev_trs R)" (is "_ \<in> rstep ?S")
proof -
  from rstep_imp_C_s_r[OF step] obtain C \<sigma> l r where rule: "(l,r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  hence rrule: "(rev_term \<box> l, rev_term \<box> r) \<in> ?S" unfolding rev_trs_def by auto
  from string_s[simplified s] have uC: "unary_ctxt C" and uls: "unary_term (l \<cdot> \<sigma>)"
    using unary_ctxt_appl_imp_unary_term[of C "l \<cdot> \<sigma>"] unary_ctxt_appl_imp_unary_ctxt[of C "l \<cdot> \<sigma>"] by auto
  from uls have ul: "unary_term l" by (rule unary_sub_appl_imp_unary_term)
  from string_t[simplified t] have urs: "unary_term (r \<cdot> \<sigma>)"
    by (rule unary_ctxt_appl_imp_unary_term)
  hence ur: "unary_term r" by (rule unary_sub_appl_imp_unary_term)
  let ?sig = "sub (get_var (l \<cdot> \<sigma>)) (rev_ctxt \<box> C)"
  let ?xsig = "get_sub \<sigma> (get_var l)"
  let ?sig2 = "Sub (\<lambda> x. Var (get_var ?xsig))"
  let ?revl = "rev_ctxt \<box> (term_to_ctxt l)"
  let ?revr = "rev_ctxt \<box> (term_to_ctxt r)"
  let ?C = "(rev_ctxt \<box> (term_to_ctxt ?xsig)) \<cdot>c ?sig"
  have url: "unary_ctxt ?revl" using ul by (simp add: unary_rev_ctxt, rule unary_term_imp_unary_ctxt)
  have urr: "unary_ctxt ?revr" using ur by (simp add: unary_rev_ctxt, rule unary_term_imp_unary_ctxt)
  have uxsig: "unary_term ?xsig" by (rule unary_sub_appl_imp_unary_sub[OF uls])
  have "rev_term \<box> s = rev_term \<box> C\<langle>l \<cdot> \<sigma>\<rangle>" by (simp add: s)
  also have "\<dots> = (rev_term \<box> (l \<cdot> \<sigma>)) \<cdot> ?sig" by (rule rev_term_ctxt_apply[OF uC uls])
  also have "\<dots> = (rev_term ?revl ?xsig) \<cdot> ?sig" using rev_term_sub_apply[OF ul, of \<box> \<sigma>] by simp
  also have "\<dots> = (rev_ctxt \<box> (term_to_ctxt ?xsig) \<circ>c ?revl)\<langle>Var (get_var ?xsig)\<rangle> \<cdot> ?sig" by (simp only: rev_term_to_rev_ctxt[OF uxsig url]) 
  also have "\<dots> = (rev_ctxt \<box> (term_to_ctxt ?xsig))\<langle>?revl \<langle>Var (get_var ?xsig)\<rangle>\<rangle> \<cdot> ?sig" by simp
  also have "\<dots> = (rev_ctxt \<box> (term_to_ctxt ?xsig))\<langle>rev_term \<box> l \<cdot> ?sig2\<rangle> \<cdot> ?sig" by (simp only: rev_ctxt_to_rev_term[OF ul, of \<box> \<sigma>, simplified])
  also have "\<dots> = ?C \<langle>rev_term \<box> l \<cdot> ?sig2 \<cdot> ?sig\<rangle>" by simp 
  finally have idl: "rev_term \<box> s = ?C \<langle>rev_term \<box> l \<cdot> (?sig2 \<circ>s ?sig)\<rangle>"
    by (simp only: sub_compose_apply_term_distrib)
  have "rev_term \<box> t = rev_term \<box> C\<langle>r \<cdot> \<sigma>\<rangle>" by (simp add: t)
  also have "\<dots> = (rev_term \<box> (r \<cdot> \<sigma>)) \<cdot> ?sig" 
    by (simp only: rev_term_ctxt_apply[OF uC urs], simp add: get_var_sub[OF ul] get_var_sub[OF ur]
      bspec[OF same_vars rule, simplified])
  also have "\<dots> = (rev_term ?revr ?xsig) \<cdot> ?sig" using rev_term_sub_apply[OF ur, of \<box> \<sigma>] bspec[OF same_vars rule, simplified] by simp
  also have "\<dots> = (rev_ctxt \<box> (term_to_ctxt ?xsig) \<circ>c ?revr)\<langle>Var (get_var ?xsig)\<rangle> \<cdot> ?sig" by (simp only: rev_term_to_rev_ctxt[OF uxsig urr]) 
  also have "\<dots> = (rev_ctxt \<box> (term_to_ctxt ?xsig))\<langle>?revr \<langle>Var (get_var ?xsig)\<rangle>\<rangle> \<cdot> ?sig" by simp
  also have "\<dots> = (rev_ctxt \<box> (term_to_ctxt ?xsig))\<langle>rev_term \<box> r \<cdot> ?sig2\<rangle> \<cdot> ?sig" by (simp only: bspec[OF same_vars rule, simplified], simp only: rev_ctxt_to_rev_term[OF ur, of \<box> \<sigma>, simplified])
  also have "\<dots> = ?C \<langle>rev_term \<box> r \<cdot> ?sig2 \<cdot> ?sig\<rangle>" by simp
  finally have idr: "rev_term \<box> t = ?C \<langle>rev_term \<box> r \<cdot> (?sig2 \<circ>s ?sig)\<rangle>"
    by (simp only: sub_compose_apply_term_distrib)
  show ?thesis using rrule by (simp only: idl idr rstep_iff_rstep_r_c_s rstep_r_c_s_def, force)
qed

definition
  check_unary_signature :: "('f, 'v) rules \<Rightarrow> shows check"
where
  "check_unary_signature R \<equiv>
     check_all (\<lambda>(l,r). unary_term l \<and> unary_term r) R
       <+? (\<lambda>_. shows_string ''the trs does not have a unary signature'')"

lemma unary_funas_conv: 
  assumes "unary_term l"
  and "(f,i) \<in> funas_term l"
  shows "i = 1"
using assms by (induct l, auto)

lemma funas_unary_conv:
  assumes "\<forall> (f,i) \<in> funas_term l. i = 1"
  shows "unary_term l"
using assms 
proof (induct l, simp)
  case (Fun f ss)
  hence "length ss = 1" by auto
  from this obtain s where ss: "ss = [s]" by (cases ss, simp, cases "tl ss", auto)
  with Fun have s: "unary_term s" by auto
  from s ss show ?case by auto 
qed
  
lemma check_unary_signature_sound : "isOK(check_unary_signature R) = unary_sig (funas_trs (set R))" (is "_ = ?r")
proof (unfold check_unary_signature_def, simp)
  show "(\<forall> x \<in> set R. (\<lambda>(l,r). unary_term l \<and> unary_term r) x) = ?r" (is "?l = _")
  proof
    assume ?l
    show ?r unfolding unary_sig_def funas_trs_def funas_rule_def
    proof (rule ballI2)
      fix f i
      assume "(f,i) \<in> (\<Union>r \<in> set R. funas_term (fst r) \<union> funas_term (snd r))"
      then obtain r where r: "r \<in> set R" and lr: "(f,i) \<in> funas_term (fst r) \<union> funas_term (snd r)" by auto
      from lr show "i = 1"
      proof
        assume "(f,i) \<in> funas_term (fst r)"
        thus "i = 1" using `?l` unary_funas_conv[of "fst r" f i] r by auto
      next
        assume "(f,i) \<in> funas_term (snd r)"
        thus "i = 1" using `?l` unary_funas_conv[of "snd r" f i] r by auto
      qed
    qed
  next
    assume ?r
    show ?l (is "\<forall> x \<in> set R. ?P x")
    proof 
      fix lr
      assume lr: "lr \<in> set R"
      show "?P lr"
      proof (cases lr)
        case (Pair l r)
        with `?r` lr have l: "\<forall> (f,i) \<in> funas_term l. i = 1" and r: "\<forall> (f,i) \<in> funas_term r. i = 1"
          unfolding unary_sig_def funas_trs_def funas_rule_def by auto       
        have "unary_term l \<and> unary_term r" using funas_unary_conv[OF l] funas_unary_conv[OF r] ..
        with Pair show ?thesis by auto
      qed
    qed
  qed
qed
   
definition
  check_string_rev :: "('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows check"
where
  "check_string_rev R rR \<equiv> do {
     check_unary_signature R;
     check_subseteq (map rev_rule R) rR
       <+? (\<lambda>rule. ''the rule '' +#+ shows_rule rule +@+ '' is missing'' +#+ shows_nl)
   }"     

lemma vars_term_unary_is_get_var: assumes "unary_term t" shows "vars_term t = {get_var t}"
using assms by (induct rule: unary_term_induct, auto)

lemma unary_rev_term: fixes t :: "('f,'v)term" assumes "unary_term t" shows "unary_term (rev_term \<box> t)"
proof -
  {
    fix C :: "('f,'v)ctxt"
    assume "unary_ctxt C"
    with assms have "unary_term (rev_term C t)"
    proof (induct arbitrary: C rule: unary_term_induct)
      case (Fun f s C) thus ?case by auto
    next
      case (Var x C) thus ?case by (simp, induct rule: unary_ctxt_induct, auto)
    qed
  }
  thus ?thesis by auto
qed

lemma unary_sig_to_unary_term: assumes F: "unary_sig F"
  and t: "t \<in> wf_terms F"
  shows "unary_term t"
using t proof (induct t)
  case (Var x) thus ?case by simp
next
  case (Fun f ss)
  with F have "length ss = 1" unfolding unary_sig_def by auto
  from this obtain s where ss: "ss = [s]" by (cases ss, simp, cases "tl ss", auto)
  with Fun show ?case by auto
qed

lemma string_rev_sound_tt_impl: "sound_tt_impl (\<lambda> R rR. isOK (check_string_rev R rR))"
unfolding sound_tt_impl_def
proof (intro allI impI)
  fix R rR
  assume check: "isOK(check_string_rev R rR)"
  and SN: "SN (rstep (set rR))"
  let ?F = "funas_trs (set R)"
  from check have 
    subset: "rev_rule ` set R \<subseteq> set rR" and 
    unary_sig: "unary_sig ?F"
    unfolding check_string_rev_def by (auto simp: check_unary_signature_sound)
  from subset have rev_trs: "rev_trs (set R) \<subseteq> set rR" unfolding rev_trs_def by auto
  have SN2: "SN (rstep (rev_trs (set R)))" by (rule SN_subset[OF SN rstep_mono[OF rev_trs]])
  {
    fix l r
    assume lr: "(l,r) \<in> set R"
    with SN_rstep_imp_wf_trs[OF SN2] have vars: "vars_term (rev_term \<box> r) \<subseteq> vars_term (rev_term \<box> l)" unfolding wf_trs_def rev_trs_def  by auto
    from lr check have ul: "unary_term l" and ur: "unary_term r" unfolding check_string_rev_def check_unary_signature_def by auto
    from vars_term_unary_is_get_var[OF unary_rev_term[OF ul]] get_var_rev_term[OF ul, of \<box>, simplified] 
         vars_term_unary_is_get_var[OF unary_rev_term[OF ur]] get_var_rev_term[OF ur, of \<box>, simplified]
         vars have "{get_var r} \<subseteq> {get_var l}" by auto
    hence "get_var l = get_var r" by auto
  }
  hence same_vars: "\<forall> (l,r) \<in> set R. get_var l = get_var r" by auto
  show "SN (rstep (set R))"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    hence "\<not> (SN (wfrstep (set R)))" using SN_wfrstep_SN_rstep_conv by auto   
    from this obtain f where steps: "\<forall> n. (f n, f (Suc n)) \<in> wfrstep (set R)" unfolding SN_defs by auto
    let ?g = "\<lambda> n. rev_term \<box> (f n)"
    have steps2: "\<forall> n. (?g n, ?g (Suc n)) \<in> rstep (rev_trs (set R))"
    proof
      fix n
      from spec[OF steps, of n]
      have 
        step: "(f n, f (Suc n)) \<in> rstep (set R)" and 
        sign: "f n \<in> wf_terms ?F"
        unfolding wfrstep_def by auto
      from spec[OF steps, of "Suc n"]
      have sigsn: "f (Suc n) \<in> wf_terms ?F"
        unfolding wfrstep_def by auto
      have ufn: "unary_term (f n)" by (rule unary_sig_to_unary_term[OF unary_sig sign])
      have ufsn: "unary_term (f (Suc n))" by (rule unary_sig_to_unary_term[OF unary_sig sigsn])
      from rstep_imp_rev_rstep[OF ufn ufsn same_vars step]
      show "(?g n, ?g (Suc n)) \<in> rstep (rev_trs (set R))" .
    qed
    with SN2 show False unfolding SN_defs by auto
  qed
qed

lemma rev_term_rev_term_id : 
  assumes "unary_term t"
  shows "rev_term \<box> (rev_term \<box> t) = t" (is "rev_term ?box _ = _")
using assms 
proof (induct t rule: unary_term_induct, simp)
  case (Fun f t)
  have ub: "unary_ctxt ?box" by simp
  let ?C = "More f [] \<box> []" 
  let ?rev = "rev_term \<box>"
  have ur: "unary_term (?rev t)" by (rule unary_rev_term[OF Fun(1)])
  have var: "get_var (?rev t) = get_var t" using get_var_rev_term[OF Fun(1)] by simp
  have "?rev (?rev (Fun f [t])) = 
    ?rev (rev_term ?C t)" by simp
  also have "\<dots> = ?rev (rev_term ?C \<box>\<langle>t\<rangle>)" by simp
  also have "\<dots> = ?rev (?rev t \<cdot> sub (get_var t) ?C)" unfolding rev_term_ctxt_apply[OF ub Fun(1), of ?C] by simp
  also have "\<dots> = Fun f [(rev_ctxt \<box> (term_to_ctxt (?rev t)))\<langle>Var (get_var t)\<rangle>]" unfolding rev_term_sub_apply[OF ur ub] by (simp add: sub_def var)
  also have "\<dots> = Fun f [?rev (?rev t)]" using rev_term_to_rev_ctxt[symmetric, OF ur ub, simplified, simplified var] by simp
  also have "\<dots> = Fun f [t]" using Fun(2) by simp
  finally show ?case .
qed


definition
  check_string_rev_nt :: "('f::show, 'v::show) rules \<Rightarrow> ('f, 'v) rules \<Rightarrow> shows check"
where
  "check_string_rev_nt R rR \<equiv> do {
     check_unary_signature R;
     check_subseteq rR (map rev_rule R)
       <+? (\<lambda>rule. ''the rule '' +#+ shows_rule rule +@+ '' is not allowed'' +#+ shows_nl)
   }"

lemma check_string_rev_nt_sound: assumes check: "isOK(check_string_rev_nt R rR)" and nSN: "\<not> SN (rstep (set rR))"
  shows "\<not> SN (rstep (set R))"
proof
  assume SN: "SN (rstep (set R))"
  from check have subset: "set rR \<subseteq> rev_rule ` (set R)" and unary: "\<And> l r. (l,r) \<in> set R \<Longrightarrow> unary_term l \<and> unary_term r" unfolding check_string_rev_nt_def 
    by (auto simp: check_unary_signature_def)
  have unary_rR: "\<And> l r. (l,r) \<in> set rR \<Longrightarrow> unary_term l \<and> unary_term r"
  proof -
    fix l r
    assume "(l,r) \<in> set rR"
    with subset obtain ll rr where mem: "(ll,rr) \<in> set R" and lr: "(l,r) = rev_rule (ll,rr)" by auto
    from unary[OF mem] unary_rev_term[of ll] unary_rev_term[of rr] lr show "unary_term l \<and> unary_term r" by auto
  qed
  hence check_unary: "isOK(check_unary_signature rR)" unfolding check_unary_signature_def by auto
  have "rev_rule ` (set rR) \<subseteq> rev_rule ` (rev_rule ` (set R))" using subset by auto
  also have "\<dots> = set R"
  proof -
    {
      fix l r
      assume lr: "(l,r) \<in> set R"
      with unary have ul: "unary_term l" and ur: "unary_term r" by auto
      from rev_term_rev_term_id[OF ul] rev_term_rev_term_id[OF ur]
      have "rev_term \<box> (rev_term \<box> l) = l" and "rev_term \<box> (rev_term \<box> r) = r" by auto
    }
    thus ?thesis by force
  qed
  finally have subset: "rev_rule ` (set rR) \<subseteq> set R" by auto
  from string_rev_sound_tt_impl[unfolded sound_tt_impl_def] SN nSN have "\<not> isOK(check_string_rev rR R)" by auto
  thus False
    by (unfold check_string_rev_def, simp add: check_unary subset)
qed

end
